using System;

namespace Indexer.Engine
{
	public class Highlighter
	{
		public Highlighter()
		{
		}
		
		public static char[] HtmlSpecialChars
		{
			get
			{
				return new char[]{'"', '&', '<', '>'};
			}
		}
		
		public static string HtmlDecode(string htmlText)
		{
			return htmlText.Replace("&quot;", "\"").Replace("&lt;", "<").Replace("&gt;", ">").Replace("&amp;", "&");
		}

		private static string SpaceEncode(string htmlText, bool preserveSpaces)
		{
			if ((!preserveSpaces) || String.IsNullOrEmpty(htmlText))
			{
				return htmlText;
			}
		
			System.Text.StringBuilder buffer = new System.Text.StringBuilder();
			int head = 0;
			while (true)
			{
				int i = htmlText.IndexOf("  ", head);
				if (i < 0)
				{
					// no more double spaces
					buffer.Append(htmlText.Substring(head));
					break;
				}
				else
				{
					buffer.Append(htmlText.Substring(head, i - head));
					int len = 0;
					while ((i + len < htmlText.Length) && (htmlText[i + len] == ' '))
					{
						len++;
					}
					// now we have "len" consecutive spaces at position "i"
					for(int counter = 0; counter < len; counter++)
					{
						buffer.Append("&nbsp;");
					}
					head = i + len;
				}
			}
			return buffer.ToString();
		}
		
		public static string HtmlEncode(string plainText)
		{
			if (plainText == null || plainText.Length == 0)
			{
				return "";
			}
			
			System.Text.StringBuilder result = new System.Text.StringBuilder(plainText.Length);
			
			for (int index = 0; index < plainText.Length; index++)
			{
				char ch = plainText[index];
				
				switch (ch)
				{
					case '"': 
						result.Append("&quot;");
						break;
					
					case '&': 
						result.Append("&amp;");
						break;
					
					case '<': 
						result.Append("&lt;");
						break;
					
					case '>': 
						result.Append("&gt;");
						break;
					
					default: 
						if (ch < 128)
						{
							result.Append(ch);
						}
						else
						{
							result.Append("&#").Append((int) ch).Append(";");
						}
						break;
					
				}
			}
			
			return result.ToString();
		}
		
		public static void FormatDocument(string sourceFileName, string targetFileName)
		{
			InternalFormatDocument(null, sourceFileName, null, targetFileName, true, false, new System.Threading.NullProgress());
		}
		
		private static long ConvertTabsToSpaces(System.IO.TextReader sourceStream, string sourceFileName, string targetFileName, System.Text.Encoding targetEncoding)
		{
			System.IO.TextReader r;
			
			if (sourceStream == null)
			{
				r = System.Text.FileStream.CreateReader(sourceFileName);
			}
			else
			{
				r = sourceStream;
			}
			
			long rowCount = 0;
			try
			{
				System.IO.TabbedTextWriter w = new System.IO.TabbedTextWriter(targetFileName, false, targetEncoding);
				w.TabSize.OutputMode = System.IO.TabbedTextWriter.Options.Conversion.TabsToSpaces;
				try
				{
					string line;
					while ((line = r.ReadLine()) != null)
					{
						w.WriteLine(line);
						rowCount++;
					}
				}
				finally
				{
					w.Dispose();
				}
			}
			finally
			{
				if (sourceStream == null)
				{
					r.Dispose();
				}
			}
			return rowCount;
		}
		
		private static void InternalFormatDocument(System.IO.TextReader sourceStream, string sourceFileName, Options formatOptions, string targetFileName, bool includeHeaders, bool lineNumbers, System.Threading.IProgress progress)
		{
			string options = "";
			if (!includeHeaders)
			{
				options = "-dh";
			}
			if (lineNumbers)
			{
				options = options + " -ln";
			}
			
			bool useColorer = (formatOptions == null) ? true : formatOptions.Colorize;
			
			sourceFileName = System.IO.Path.GetFullPath(sourceFileName);
			
			string tempFile = System.IO.Path.GetTempFileName();
			System.IO.File.Delete(tempFile);
			string extension;
			if ((formatOptions != null) && !String.IsNullOrEmpty(formatOptions.Extension))
			{
				extension = formatOptions.Extension;
			}
			else
			{
				extension = System.IO.Path.GetExtension(sourceFileName);
			}
			tempFile = tempFile + extension;

			bool createNormalizedSource = (formatOptions == null) ? false : formatOptions.CreateNormalizedSourceFile;
			
			if (createNormalizedSource)
			{
				formatOptions.NormalizedSourceFile = tempFile;
			}
			
			try
			{
				// TODO: show progress for conversion
				long rowCount = ConvertTabsToSpaces(sourceStream, sourceFileName, tempFile, System.Text.Encoding.UTF8);
				
				if (createNormalizedSource)
				{
					formatOptions.NormalizedSourceFileRowCount = rowCount;
				}
				
				sourceFileName = tempFile;
				
				targetFileName = System.IO.Path.GetFullPath(targetFileName);
				
				if (useColorer)
				{
					progress.Open(0, (ulong)(new System.IO.FileInfo(sourceFileName)).Length * 7 /* TODO: TUNE: heuristics */);
					try
					{
						System.Diagnostics.Process runner = new System.Diagnostics.Process();
						try
						{
							runner.StartInfo.Arguments = String.Format("-h -ei UTF-8 -eo UTF-8 -dc {0} \"{1}\" -o \"{2}\"", options, sourceFileName, targetFileName);
							runner.StartInfo.FileName = System.IO.Path.Combine(System.IO.Path.GetDirectoryName(System.Reflection.Assembly.GetEntryAssembly().Location), "colorer.exe");
							runner.StartInfo.UseShellExecute = false;
							runner.StartInfo.CreateNoWindow = true;
							// TODO: check results and throw exception
							runner.Start();
							DateTime startTime = DateTime.Now;
							long lastLength = -1;
							while (!runner.HasExited)
							{
								if (DateTime.Now.Subtract(startTime).TotalSeconds > 20 /* TODO: TUNE: how many seconds to wait */)
								{
									// formatting process is hung-up. kill it and pretend that nothing happened
									// by using an alternate way - no colorizing
									runner.Kill();
									useColorer = false;
								}
								System.IO.FileInfo info = new System.IO.FileInfo(targetFileName);
								if (info.Exists)
								{
									long newLength = info.Length;
									progress.Step((ulong)newLength);
									if (newLength != lastLength)
									{
										lastLength = newLength;
										// reset time-out timer
										startTime = DateTime.Now;
									}
								}
								if (runner.WaitForExit(500 /* TODO: TUNE: */))
								{
									break;
								}
							}
							runner.WaitForExit();
						}
						finally
						{
							runner.Dispose();
						}
					}
					finally
					{
						progress.Close();
					}
				}
				
				// handle the case when colorer is not needed or crashed
				if (!useColorer)
				{
					// TODO: use progress
					System.IO.TextReader input = System.Text.FileStream.CreateReader(tempFile);
					try
					{
						System.IO.TextWriter output = new System.IO.StreamWriter(targetFileName, false, System.Text.Encoding.UTF8);
						try
						{
							int lineNumber = 0;
							string line;
							int lineNumberWidth = 1;
							{
								long beyondLimit = 10;
								while (rowCount >= beyondLimit)
								{
									lineNumberWidth++;
									beyondLimit *= 10;
								}
							}
							while ((line = input.ReadLine()) != null)
							{
								if (lineNumbers)
								{
									string prefix = lineNumber.ToString();
									while (prefix.Length < lineNumberWidth)
									{
										prefix = " " + prefix;
									}
									output.Write(prefix + ": ");
								}
								output.WriteLine(HtmlEncode(line));
								lineNumber++;
							}
						}
						finally
						{
							output.Close();
						}
					}
					finally
					{
						input.Close();
					}
				}
			}
			finally
			{
				if (!createNormalizedSource)
				{
					System.IO.File.Delete(tempFile);
				}
			}
		}
		
		private enum TokenType
		{
			SpanOpen,
			SpanClose,
			Text
		}
		
		private class Token
		{
			public string Value;
			public TokenType Type;
		}
		
		private class SpanTokenizer: System.Collections.Generic.IEnumerator<Token>, System.Collections.Generic.IEnumerable<Token>
		{
			private string _line;
			private int _head;
			public SpanTokenizer(string line)
			{
				_line = line;
				Reset();
			}
			
			private Token _current;
			public Token Current 
			{
				get 
				{
					return _current;
				}
			}
			
			object System.Collections.IEnumerator.Current 
			{
				get 
				{
					return _current;
				}
			}
			
			public void Dispose()
			{
			}
			
			public bool MoveNext()
			{
				if (_head >= _line.Length)
				{
					return false;
				}
				int i = _line.IndexOf('<', _head);
				if (i < 0)
				{
					_current = new Token();
					_current.Type = TokenType.Text;
					_current.Value = _line.Substring(_head);
					_head += _current.Value.Length;
					return true;
				}
				else
				{
					if (i > _head)
					{
						_current = new Token();
						_current.Type = TokenType.Text;
						_current.Value = _line.Substring(_head, i - _head);
						_head += _current.Value.Length;
						return true;
					}
					else
					{
						int j = _line.IndexOf('>', i);
						_current = new Token();
						_current.Value = _line.Substring(i, j - i + 1);
						if (_current.Value.StartsWith("<span"))
						{
							_current.Type = TokenType.SpanOpen;
						}
						else if (_current.Value.StartsWith("</span"))
						{
							_current.Type = TokenType.SpanClose;
						}
						else
						{
							_current.Type = TokenType.Text;
						}
						_head += _current.Value.Length;
						return true;
					}
				}
			}
			
			public void Reset()
			{
				_head = 0;
			}
			
			public System.Collections.Generic.IEnumerator<Token> GetEnumerator()
			{
				return this;
			}
			
			System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
			{
				return this;
			}
		}
		
		private class SpanCode
		{
			private bool _encode;
			private int _offset = 0;
			public string Text;
			public System.Collections.Generic.Dictionary<int, string> Insertions;
			
			public SpanCode(string line, bool encode)
			{
				_encode = encode;
				Text = "";
				Insertions = new  System.Collections.Generic.Dictionary<int, string>();
				
				for(;;)
				{
					int i = line.IndexOf('<');
					if (i < 0)
					{
						if (_encode)
						{
							Text = Text + HtmlDecode(line);
						}
						else
						{
							Text = Text + line;
						}
						break;
					}
					int j = line.IndexOf('>', i + 1);
					if (j < 0)
					{
						// this means that input is not valid - try to recover
						// by ignoring the incomplete tag
						line = line.Substring(0, i);
						continue;
					}
					int len = j - i + 1;
					if ((i == 0) && Insertions.ContainsKey(Text.Length))
					{
						// insertion already added, expand it
						Insertions[Text.Length] = Insertions[Text.Length] + line.Substring(i, len);
					}
					else
					{
						if (_encode)
						{
							Text = Text + HtmlDecode(line.Substring(0, i));
						}
						else
						{
							Text = Text + line.Substring(0, i);
						}
						Insertions.Add(Text.Length, line.Substring(i, len));
					}
					line = line.Remove(0, i + len);
				}
			}
			
			public string Remove(int count, bool appendInsertions)
			{
				string result = "";
				while (count > 0)
				{
					result = result + RemoveOne();
					count--;
				}
				if (appendInsertions)
				{
					result = result + RemoveInsertion();
				}
				return result;
			}
			
			public string Remove()
			{
				return Remove(Text.Length, true);
			}
			
			private string RemoveInsertion()
			{
				string result = "";
				if (Insertions.TryGetValue(_offset, out result))
				{
					Insertions[_offset] = "";
				}
				return result;
			}
			
			private string RemoveOne()
			{
				string result = RemoveInsertion();
				if (Text.Length > 0)
				{
					if (_encode)
					{
						result = result + HtmlEncode(Text.Substring(0, 1));
					}
					else
					{
						result = result + Text.Substring(0, 1);
					}
					Text = Text.Remove(0, 1);
				}
				_offset++;
				return result;
			}
		}
		
		private static bool IsBlank(string line)
		{
			for(int i = 0; i < line.Length; i++)
			{
				if (line[i] != ' ')
				{
					return false;
				}
			}
			return true;
		}
		
		public class Strings: System.Collections.Generic.List<string>
		{
		}

		public class Options
		{
			// inputs
			public string Extension = null;
			public bool WordWrap = false;
			public bool UseLineNumbers = true;
			public bool Colorize = true;
			public bool PreserveSpaces = true;
			public bool CreateNormalizedSourceFile = false;
			public bool IncludeHeader = true;
			public bool TrimEndBlankLines = true;
			
			// outputs
			public string NormalizedSourceFile;
			public long NormalizedSourceFileRowCount;
		}
		
		public static Strings FormatDocument(string sourceFileName, Options formatOptions, IMatchMarker marker, string targetFileName, Strings linePreviews, System.Threading.IProgress progress)
		{
			System.IO.TextWriter target = new System.IO.StreamWriter(targetFileName, false, System.Text.Encoding.UTF8);
			try
			{
				return FormatDocument(sourceFileName, formatOptions, marker, target, linePreviews, progress);
			}
			finally
			{
				target.Dispose();
			}
		}
		
		public static Strings FormatDocument(System.IO.TextReader sourceStream, string sourceFileName, Options formatOptions, IMatchMarker marker, string targetFileName, Strings linePreviews, System.Threading.IProgress progress)
		{
			System.IO.TextWriter target = new System.IO.StreamWriter(targetFileName, false, System.Text.Encoding.UTF8);
			try
			{
				return FormatDocument(sourceStream, sourceFileName, formatOptions, marker, target, linePreviews, progress);
			}
			finally
			{
				target.Dispose();
			}
		}
		
		public struct Match
		{
			public int Row;
			public int Column;
			public int Length;
		}
		
		public interface IMatchMarker
		{
			Match[] GetHighlightPoints(string text);
			string[] Phrases {get;}
		}
		
		public static Strings FormatDocument(string sourceFileName, Options formatOptions, IMatchMarker marker, System.IO.TextWriter target, Strings linePreviews, System.Threading.IProgress progress)
		{
			return FormatDocument(null, sourceFileName, formatOptions, marker, target, linePreviews, progress);
		}

		public class PageWriter
		{
			private System.IO.TextWriter _target;
			
			private string beforeLineNumberCode = "";
			private string afterLineNumberCode = "";
			private string beforeValueCode = "";
			private string afterValueCode = "";
			private string emptyValueCode = "";
			private string lineNumberPaddingCode = "";
			private string beforeListCode = "";
			private string afterListCode = "";
			private bool wordWrap;
			private bool useLineNumbers;
			private bool preserveSpaces;
			private bool useTable;
			private bool includeHeader;
			
			public PageWriter(System.IO.TextWriter target, Options formatOptions)
			{
				_target = target;	

				wordWrap = formatOptions.WordWrap;
				useLineNumbers = formatOptions.UseLineNumbers;
				preserveSpaces = formatOptions.PreserveSpaces;
				useTable = useLineNumbers && wordWrap;
				includeHeader = formatOptions.IncludeHeader;

				if (includeHeader)
				{
					if (useTable)
					{
						// TODO: use font-size equivalent to <font size=-1>
						// beforeListCode = "<table style='font-family: Courier New,Courier,monospace; font-size: 1em;'>";
						beforeListCode = "<table style='font-family: Courier New,Courier,monospace;' border=0 cellpadding=0 cellspacing=0>";
						afterListCode = "</table>";
					}
					else if (wordWrap)
					{
						beforeListCode = "<font face=\"Courier New\" size=-1>";
						afterListCode = "</font>";
					}
					else
					{
						beforeListCode = "<pre>";
						afterListCode = "</pre>";
					}
				}
				
				if (wordWrap)
				{
					lineNumberPaddingCode = "&nbsp;";
				}
				else
				{
					lineNumberPaddingCode = " ";
				}
				
				if (useTable)
				{
					// Don't use <font size=..> but inherit if in the style from <table> element
					beforeLineNumberCode = "<tr><td valign=top><font size=-1>";
					afterLineNumberCode = "</font></td>";
					beforeValueCode = "<td valign=top><font size=-1>";
					afterValueCode = "</font></td></tr>";
					emptyValueCode = beforeValueCode + "&nbsp;" + afterValueCode;
				}
				else if (wordWrap)
				{
					afterValueCode = "<br>" + target.NewLine;
					emptyValueCode = afterValueCode;
				}
				else
				{
					afterValueCode = target.NewLine;
					emptyValueCode = afterValueCode;
				}
			}
			
			public void WriteHeader(string title)
			{
				if (includeHeader)
				{
					_target.Write("<html><head><meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\"><title>");
					_target.Write(HtmlEncode(title));
					_target.Write("</title></head><body style='color:#000000; background:#ffffff; '>");
					_target.WriteLine(beforeListCode);
				}
			}
			
			public void WriteFooter()
			{
				if (includeHeader)
				{
					_target.Write(afterListCode);
					_target.WriteLine("</body></html>");
				}
			}
			
			public string FormatLineNumber(string number, int width)
			{
				for(int prefixWidth = number.Length; prefixWidth < width; prefixWidth++)
				{
					number = lineNumberPaddingCode + number;
				}
				return number;
			}
			
			public string GetEmptyLineNumberCode(string number)
			{
				if (useLineNumbers)
				{
					return beforeLineNumberCode + number + afterLineNumberCode;
				}
				else
				{
					return "";
				}
			}

			public string GetLineNumberCode(string number)
			{
				if (useLineNumbers)
				{
					return beforeLineNumberCode + number + ":" + lineNumberPaddingCode + afterLineNumberCode;
				}
				else
				{
					return "";
				}
			}
			
			public string GetEmptyValueCode()
			{
				return emptyValueCode;
			}
			
			public string GetBeforeValueCode()
			{
				return beforeValueCode;
			}
			
			public string GetAfterValueCode()
			{
				return afterValueCode;
			}
			
			public string GetTextCode(string text)
			{
				return SpaceEncode(text, preserveSpaces);
			}
		}
		
		public static Strings FormatDocument(System.IO.TextReader sourceStream, string sourceFileName, Options formatOptions, IMatchMarker marker, System.IO.TextWriter target, Strings linePreviews, System.Threading.IProgress progress)
		{
			System.Threading.SlicedProgress overallProgress = new System.Threading.SlicedProgress(progress);
			// TODO: TUNE:
			overallProgress.Add(100);	// coloring stage
			overallProgress.Add(80);	// highlighting stage
			try
			{
				if (formatOptions == null)
				{
					formatOptions = new Options();
				}
				
				const bool singleRowBookmark = true;	// TODO: add it to arguments once fully supported

				Strings result = new Strings();
				string tempFile = System.IO.Path.GetTempFileName();
				try
				{
					InternalFormatDocument(sourceStream, sourceFileName, formatOptions, tempFile, false, true, overallProgress[0]);
	
					System.Threading.IProgress p = overallProgress[1];
					
					System.IO.Stream input = new System.IO.FileStream(tempFile, System.IO.FileMode.Open);
					p.Open(0, (ulong)input.Length);
					try
					{
						const string HitFormat = "hit{0}";
						PageWriter w = new PageWriter(target, formatOptions);
						System.IO.TextReader r = System.Text.FileStream.CreateReader(input);
						try
						{
							w.WriteHeader(sourceFileName);
							
							int row = -1;
							string line;

							bool encoderNeeded = false;
							if (marker != null)
							{
								foreach(string phrase in marker.Phrases)
								{
									if (phrase.IndexOfAny(HtmlSpecialChars) >= 0)
									{
										encoderNeeded = true;
										break;
									}
								}
							}
							
							System.Text.StringBuilder blanks = null;
							
							while ((line = r.ReadLine()) != null)
							{
								p.Step((ulong)input.Position);
								
								row++;
							
								System.Text.StringBuilder lineBuffer = new System.Text.StringBuilder();
								
								{
									// extract and increment line number. colorer uses 0-based lines
									int i = line.IndexOf(": ");
									int originalRow = Int32.Parse(line.Substring(0, i).Trim());
									if (originalRow != row)
									{
										throw new ArgumentException("Unexpected line number");
									}
									line = line.Remove(0, i + 2);
									string prefix = (row + 1).ToString();
									prefix = w.FormatLineNumber(prefix, i);
									
									if (IsBlank(line))
									{
										if (blanks == null)
										{
											blanks = new System.Text.StringBuilder();
										}
										blanks.Append(w.GetEmptyLineNumberCode(String.Format("<span style='color:#C0C0C0'>{0}:</span>", prefix)));
										blanks.Append(w.GetEmptyValueCode());
										continue;
									}
									else
									{
										if (blanks != null)
										{
											target.Write(blanks.ToString());
											blanks = null;
										}
										lineBuffer.Append(w.GetLineNumberCode(prefix));
									}
								}
								
								lineBuffer.Append(w.GetBeforeValueCode());
								
								SpanCode code = new SpanCode(line, encoderNeeded);
								Match[] marks;
								if (marker == null)
								{
									marks = null;
								}
								else
								{
									marks = marker.GetHighlightPoints(code.Text);
								}
								int markIndex = 0;
								bool useMarks = (marks != null) && (marks.Length > 0);
								
								int cut = 0;
								bool rowHit = false;
								if (useMarks)
								{
									while (useMarks)
									{
										int column = marks[markIndex].Column - cut;
										lineBuffer.Append(w.GetTextCode(code.Remove(column, true)));
										lineBuffer.Append("<span style='color:#000000; background:#ffff00; '>");
										if (!rowHit)
										{
											// TODO: bookmarks are not unique in non-singleRowBookmark mode
											string bookmark = String.Format(HitFormat, row + 1);
											lineBuffer.Append("<a name=" + bookmark + "></a>");
											result.Add(bookmark);
											if (singleRowBookmark)
											{
												rowHit = true;
											}
										}
										{
											/* we have to modify phrase if it includes embedded span-formatting:
											 * 
											 * 1. a</span>b -> a</span></span><span ourstyle>b
											 * 2. a<span..>b -> a</span><span..><span ourstyle>b
											 */
											foreach(Token token in new SpanTokenizer(code.Remove(marks[markIndex].Length, false)))
											{
												switch (token.Type)
												{
													case TokenType.SpanOpen:
													case TokenType.SpanClose:
														lineBuffer.Append("</span>");
														lineBuffer.Append(w.GetTextCode(token.Value));
														lineBuffer.Append("<span style='color:#000000; background:#ffff00; '>");
														break;
														
													default:
														lineBuffer.Append(w.GetTextCode(token.Value));
														break;
												}
											}
										}
										lineBuffer.Append("</span>");
										cut += column + marks[markIndex].Length;
										markIndex++;
										if (markIndex >= marks.Length)
										{
											useMarks = false;
										}
									}
									lineBuffer.Append(w.GetTextCode(code.Remove()));
									if (linePreviews != null)
									{
										string linePreview = lineBuffer.ToString();

										// remove tags that were added above and are valid only in the text part
										foreach(string tag in new string[]{"<tr>", "</tr>", "<td valign=top>", "</td>", "<br>", "<font size=-1>", "</font>"})
										{
											linePreview = linePreview.Replace(tag, "");
										}

										// remove bookmarks
										while(true)
										{
											int i = linePreview.IndexOf("<a");
											if (i < 0)
											{
												break;
											}
											int j = linePreview.IndexOf("a>", i + 2);
											linePreview = linePreview.Remove(i, j - i + 2);
										}
										linePreviews.Add(linePreview);
									}
								}
								else
								{
									lineBuffer.Append(w.GetTextCode(line));
								}
								lineBuffer.Append(w.GetAfterValueCode());
								target.Write(lineBuffer.ToString());
							}

							if (!formatOptions.TrimEndBlankLines)
							{
								if (blanks != null)
								{
									target.Write(blanks.ToString());
									blanks = null;
								}
							}
							
							w.WriteFooter();
						}
						finally
						{
							r.Dispose();
						}
						if (result.Count == 0)
						{
							return null;
						}
						else
						{
							return result;
						}
					}
					finally
					{
						input.Dispose();
						p.Close();
					}
				}
				finally
				{
					System.IO.File.Delete(tempFile);
				}
			}
			finally
			{
				overallProgress.Close();
			}
		}
	}
}
