using System;
using System.Runtime.InteropServices;
using System.Collections.Generic;
using System.Drawing;
using System.Windows.Forms;
using System.Configuration;
using Indexer.Engine;
using DocumentHandle = System.Int32;

namespace Indexer.Gui
{
	public partial class MainForm : Form, IIndexObserver
	{
		private Worker _worker = null;		// worker thread
		private Index _index = null;		// current index
		private SearchHistory _history = new SearchHistory();
		private Search _search = null;		// current entry
		private System.Windows.Forms.MessageBoardClient _synchronizer = null;
		
		private System.Collections.Generic.List<string> _temporaryFiles = new System.Collections.Generic.List<string>();
		
		public MainForm()
		{
			//
			// The InitializeComponent() call is required for Windows Forms designer support.
			//
			InitializeComponent();

			PrepareControls();
			
			_worker = new Worker(this);
			UpdateIndexControls();
		}

		private struct SearchModeEntry
		{
			public string ListValue;
		}
		
		private class SearchModeMap: System.Collections.Generic.Dictionary<Index.SetCombineMode, SearchModeEntry>
		{
			public void Add(Index.SetCombineMode mode, string listValue)
			{
				SearchModeEntry entry = new SearchModeEntry();
				entry.ListValue = listValue;
				base.Add(mode, entry);
			}
		}

		private Index.SetCombineMode _searchMode = Index.SetCombineMode.None;
		private Index.SetCombineMode SearchMode
		{
			get
			{
				return _searchMode;
			}
			set
			{
				if (_searchMode != value)
				{
					_searchMode = value;
					foreach(Index.SetCombineMode mode in _searchModeMap.Keys)
					{
						SearchModeEntry entry = _searchModeMap[mode];
						if (mode == _searchMode)
						{
							this.setOperatorList.Text = entry.ListValue;
							break;
						}
					}
				}
			}
		}
		
		private SearchModeMap _searchModeMap;
		private void PrepareControls()
		{
			this.toolStripContainer.TopToolStripPanel.Controls.Clear();
			this.toolStripContainer.TopToolStripPanel.Join(this.activityPanel, 0);
			this.toolStripContainer.TopToolStripPanel.Join(this.searchPanel, 0);
			this.toolStripContainer.TopToolStripPanel.Join(this.historyPanel, 0);
			
			_searchModeMap = new SearchModeMap();
			_searchModeMap.Add(Index.SetCombineMode.None, "");
			_searchModeMap.Add(Index.SetCombineMode.And, "and");
			_searchModeMap.Add(Index.SetCombineMode.AndNot, "sub");
			_searchModeMap.Add(Index.SetCombineMode.Or, "or");
			SearchMode = Index.SetCombineMode.None;
			
			new KeyHandler(this.fileList, this.FileListKeyDownUp);
			new KeyHandler(this.searchBox, this.SearchBoxKeyDownUp);
			new KeyHandler(this.fileMaskList, this.SearchBoxKeyDownUp);

			// TODO: add shortcut keys to menu items - for display
			// this.fileListMenuCopy.Text = this.fileListMenuCopy.Text + "\tCtrl+C";
		}
		
		private void Write()
		{
			Write(null);
		}

		private string _lastError;
		private void WriteLastError()
		{
			if (String.IsNullOrEmpty(_lastError))
			{
				this.statusError.Visible = false;
				InternalWrite(null);
			}
			else
			{
				InternalWrite(null);
				this.statusError.Text = " " + _lastError;
				this.statusError.Visible = true;
			}
		}
		
		private void InternalWrite(string message)
		{
			if (String.IsNullOrEmpty(message))
			{
				this.statusActivity.Text = "";
			}
			else
			{
				this.statusActivity.Text = message;
			}
		}
		
		private string _lastMessage;
		private void Write(string message)
		{
			_lastMessage = message;
			if (String.IsNullOrEmpty(message))
			{
				InternalWrite(message);
			}
			else
			{
				InternalWrite(message + "...");
			}
		}

		private void WriteProgress()
		{
			WriteProgress(null);
		}
		
		private void WriteProgress(string message)
		{
			if (String.IsNullOrEmpty(message))
			{
				Write(_lastMessage);
			}
			else
			{
				if (String.IsNullOrEmpty(_lastMessage))
				{
					InternalWrite(message + "...");
				}
				else
				{
					InternalWrite(_lastMessage + ", " + message + "...");
				}
			}
		}
		
		private void UpdateIndexControls()
		{
			if (_index == null)
			{
				this.searchPanel.Enabled = false;
				this.fileList.Items.Clear();
				this.fileList.Enabled = false;
			}
			else
			{
				if (!this.searchPanel.Enabled)
				{
					this.searchPanel.Enabled = true;
					this.searchBox.Focus();
				}
				this.fileList.Enabled = true;
			}
			UpdateModifiedIndicator();
			UpdateFileCount();
		}
		
		private void UpdateModifiedIndicator()
		{
			if (_index == null)
			{
				this.Text = "Indexer";
			}
			else
			{
				string mark = "";
				if (_index.Modified)
				{
					mark = "*";	
				}
				if (String.IsNullOrEmpty(_index.FileName))
				{
					this.Text = "Untitled" + mark + " - Indexer";
				}
				else
				{
					this.Text = System.IO.Path.GetFileName(_index.FileName) + mark + " - Indexer";
				}
			}
		}
		
		private void UpdateFileCount()
		{
			if (_index == null)
			{
				UpdateFileCount(0);
			}
			else
			{
				UpdateFileCount(_index.DocumentsCount);
			}
		}
		
		private void UpdateFileCount(int fileCount)
		{
			this.statusDocumentCount.Text = N2S(fileCount);
			this.statusFoundDocumentCount.AutoSize = false;
			this.statusFoundDocumentCount.Width = this.statusDocumentCount.Width;
		}

		private void UpdateFoundDocumentCount(int documentCount)
		{
			this.statusFoundDocumentCount.Text = N2S(documentCount);
		}
		
		private IniFile GetConfig(bool writable)
		{
			if (writable)
			{
				return new IniFile(System.Windows.Forms.Environment.GetFullPath("config.ini", System.Windows.Forms.UserProfile.NonRoamingUser));
			}
			else
			{
				return new IniFile(System.IO.Path.Combine(System.IO.Path.GetDirectoryName(System.Reflection.Assembly.GetEntryAssembly().Location), "indexer.ini"));
			}
		}

		private IniFile GetConfig()
		{
			return GetConfig(true);
		}

		private bool GetComparer()
		{
			string fileName;
			string arguments;
			return GetComparer(out fileName, out arguments);
		}
		
		private bool GetComparer(out string fileName, out string arguments)
		{
			IniFile config = GetConfig(false);
			try
			{
				System.Collections.Generic.List<string> comparers = new System.Collections.Generic.List<string>();
				
				foreach(string key in config.ReadSection("Comparers").AllKeys)
				{
					if (key.EndsWith(".path", StringComparison.InvariantCultureIgnoreCase))
					{
						string comparerName = key.Substring(0, key.Length - ".path".Length).ToUpper();
						if (!comparers.Contains(comparerName))
						{
							comparers.Add(comparerName);
						}
					}
				}
				
				fileName = null;
				arguments = null;
				
				foreach(string comparerName in comparers)
				{
					fileName = config.ReadString(String.Format("Comparers/{0}.path", comparerName), null);
					arguments = config.ReadString(String.Format("Comparers/{0}.arguments", comparerName), null);
					
					if (!String.IsNullOrEmpty(fileName) && !String.IsNullOrEmpty(arguments))
					{
						if (System.IO.File.Exists(fileName))
						{
							return true;
						}
					}
				}
				
				return false;
			}
			finally
			{
				config.Dispose();
			}
		}
		
		private void InternalOpenIndex()
		{
			InternalOpenIndex(null);
		}

		private void SetLastIndex(string value)
		{
			ISettingsWriter config = GetConfig();
			try
			{
				if (String.IsNullOrEmpty(value))
				{
					config.Delete("Path/LastIndex");
				}
				else
				{
					config.WriteString("Path/LastIndex", value);
				}
			}
			finally
			{
				config.Dispose();
			}
		}

		private void InternalOpenIndex(string fileName)
		{
			LockControls(InterruptMode.Enabled);
			try
			{
				OpenIndexCommand command = new OpenIndexCommand();
				command.Observer = _worker;
				
				if (String.IsNullOrEmpty(fileName))
				{
					Write("Create index");
					_worker.Execute(command);
					_index = command.Result;
					SetLastIndex(null);
				}
				else
				{
					try
					{
						Write("Open index");
						this.Text = System.IO.Path.GetFileName(fileName) + " - Indexer";
						
						// TODO: this is a hack
						_originalTitle = this.Text;
						
						command.FileName = fileName;
						_worker.Execute(command);
						_index = command.Result;
						SetLastIndex(_index.FileName);
						this.openIndexDialog.FileName = _index.FileName;
					}
					catch(Exception e)
					{
						command.FileName = null;
						_worker.Execute(command);
						_index = command.Result;
						SetLastIndex(null);
						throw e;
					}
				}
			}
			finally
			{
				ConnectMaintainer();
				
				UnLockControls(FocusMode.Auto);
				UpdateIndexControls();
				Write();
			}
		}
		
		#region Screen locking
		private int _lockCount = 0;
		private bool _searchPanelEnabled;
		private bool _historyPanelEnabled;
		private bool _fileListEnabled;
		private Cursor _oldCursor;

		private enum InterruptMode
		{
			Enabled,
			Disabled
		}
		
		private void LockControls(InterruptMode mode)
		{
			InternalLockControls(true, mode, FocusMode.Auto);
		}

		private enum FocusMode
		{
			Auto,
			Keep
		}
		
		private void UnLockControls(FocusMode mode)
		{
			InternalLockControls(false, InterruptMode.Enabled, mode);
		}
		
		public class BusyLock
		{
			System.Windows.Forms.Form _control;
			public BusyLock(System.Windows.Forms.Form control)
			{
				_control = control;
			}
	
			private int _busyLock = 0;
			public void Busy(bool activate)
			{
				if (activate)
				{
					if (_busyLock == 0)
					{
						_control.Cursor = System.Windows.Forms.Cursors.WaitCursor;
					}
					_busyLock++;
				}
				else
				{
					_busyLock--;
					if (_busyLock == 0)
					{
						_control.Cursor = System.Windows.Forms.Cursors.Default;
					}
				}
			}
		}
		
		private BusyLock _busy = null;
		private void Wait(bool activate)
		{
			if (_busy == null)
			{
				_busy = new BusyLock(this);
			}
			_busy.Busy(activate);
		}
		
		private void InternalLockControls(bool doLock, InterruptMode interruptMode, FocusMode focusMode)
		{
			if (doLock)
			{
				if (_lockCount == 0)
				{
					if (_synchronizer != null)
					{
						_synchronizer.Enabled = false;
					}
					
					_lastError = null;
					this.statusError.Visible = false;
					
					System.Windows.Forms.IconHelper.LoadIcon(this, "Gadgets", "Busy.Icon");
					this.Update();
					
					_searchPanelEnabled = this.searchPanel.Enabled;
					this.searchBox.Select(0, 0);
					this.searchPanel.Enabled = false;
					_fileListEnabled = this.fileList.Enabled;
					this.fileList.Enabled = false;
					_historyPanelEnabled = this.historyPanel.Enabled;
					this.historyPanel.Enabled = false;
					
					this.menuFile.Enabled = false;
					this.menuIndex.Enabled = false;
					// this.menuHelp.Enabled = false;
					
					_escape = false;
					_interrupted = false;

					CloseButton.SetEnabled(this, false);
					
					this.Update();
		
					_oldCursor = this.Cursor;
					this.Cursor = Cursors.AppStarting;
					
					this.interruptButton.Enabled = interruptMode == InterruptMode.Enabled;

					this.AllowDrop = false;
					
					StartProgressTimer();
				}
				_lockCount++;
			}
			else
			{
				_lockCount--;
				if (_lockCount == 0)
				{
					this.searchPanel.Enabled = _searchPanelEnabled;
					this.fileList.Enabled = _fileListEnabled;
					this.historyPanel.Enabled = _historyPanelEnabled;
					
					this.menuFile.Enabled = true;
					this.menuIndex.Enabled = true;
					this.menuHelp.Enabled = true;

					CloseButton.SetEnabled(this, true);

					this.interruptButton.Enabled = false;
					
					this.Update();
					
					this.AllowDrop = true;
					
					StopProgressTimer();
					
					this.Cursor = _oldCursor;
					
					if (this.searchBox.Enabled && (focusMode == FocusMode.Auto))
					{
						this.searchBox.Focus();
					}
					
					UpdateModifiedIndicator();
					
					System.Windows.Forms.IconHelper.LoadIcon(this, "$this.Icon");
					
					if (_synchronizer != null)
					{
						_synchronizer.Enabled = true;
					}
				}
			}
		}
		
		private bool Locked
		{
			get
			{
				return _lockCount > 0;
			}
		}
		#endregion

		private void OpenSelectedFiles()
		{
			if (this.fileList.SelectedIndices.Count == 0)
			{
				return;
			}
			if (!String.IsNullOrEmpty(this.searchBox.Text))
			{
				System.Windows.Forms.Clipboard.SetText(this.searchBox.Text);
			}
			Write("Export");
			LockControls(InterruptMode.Disabled);
			try
			{
				try
				{
					int count = fileList.SelectedIndices.Count;
					ExportFilesCommand command = new ExportFilesCommand();
					command.Colorize = new System.Collections.BitArray(count);
					command.FileNames = new string[count];
					command.Handles = new DocumentHandle[count];
					{
						int i = 0;
						foreach(int index in fileList.SelectedIndices)
						{
							command.FileNames[i] = _search.Results[index];
							switch (_search.Results.GetItemKind(index))
							{
								case Index.DocumentKind.Indexed:
								case Index.DocumentKind.Indexed | Index.DocumentKind.Cached:
									command.Colorize[i] = true;
									command.Handles[i] = _search.Results.GetItemHandle(index);
									break;
							}
							i++;
						}
					}
					command.TemporaryFiles = _temporaryFiles;
					command.Index = _index;
					command.Query = _search.Query;
					
					if (fileList.SelectedIndices.Count == fileList.VirtualListSize)
					{
						command.Title = _search.DisplayLabel;
					}
					else
					{
						command.Title = String.Format("{0}, showing {1}",
						                              _search.DisplayLabel,
						                              N2S(fileList.SelectedIndices.Count));
					}
					
					_worker.Execute(command);
	
					OpenFileCommand opener = new OpenFileCommand();
					opener.FileName = command.Result;
					_worker.Execute(opener);
				}
				catch(Exception error)
				{
					ShowException(error, this.DefaultErrorHandler);
				}
			}
			finally
			{
				WriteLastError();
				UnLockControls(FocusMode.Keep);
			}
		}
		
		void MenuQuitClick(object sender, EventArgs e)
		{
			this.Close();
		}

		bool InternalCloseIndex()
		{
			if (_index == null)
			{
				return true;
			}
			
			if (!_index.Modified)
			{
				DisconnectMaintainer();
				ClearAllResults();
				_index.Dispose();
				_index = null;
				UpdateIndexControls();
				return true;
			}
			switch (MessageBox.Show("Save changes", this.Text, MessageBoxButtons.YesNoCancel, MessageBoxIcon.Question))
			{
				case DialogResult.Yes:
					if (DoSave())
					{
						return InternalCloseIndex();
					}
					else
					{
						return false;
					}

				case DialogResult.No:
					{
						_index.Modified = false;
						return InternalCloseIndex();
					}

				default:
					return false;
			}
		}
		
		void MenuFileNewClick(object sender, EventArgs e)
		{
			try
			{
				try
				{
					if (InternalCloseIndex())
					{
						InternalOpenIndex();
					}
				}
				catch (Exception error)
				{
					ShowException(error, this.OpenIndexErrorHandler /* TODO: should be SaveIndexErrorHandler?*/);
				}
			}
			finally
			{
				WriteLastError();
			}
		}

		private void InternalAddMounts(string[] fileNames)
		{
			Write("Add archives");
			LockControls(InterruptMode.Disabled);
			try
			{
				try
				{
					ClearAllResults();
					MountArchiveCommand command = new MountArchiveCommand();
					command.Index = _index;
					command.FileNames = fileNames;
					_worker.Execute(command);
				}
				catch(Exception error)
				{
					ShowException(error, this.UpdateIndexErrorHandler);
				}
			}
			finally
			{
				WriteLastError();
				UnLockControls(FocusMode.Keep);
				UpdateFileCount();
			}
		}
	
		private bool UpdateIndexErrorHandler(ErrorCategory category, ref string message)
		{
			if (category == ErrorCategory.OperationCanceled)
			{
				message = "Index was not completely updated";
				return true;
			}
			return false;
		}

		private bool SearchErrorHandler(ErrorCategory category, ref string message)
		{
			if (category == ErrorCategory.OperationCanceled)
			{
				message = "Incomplete search results";
				return true;
			}
			return false;
		}

		private bool OpenIndexErrorHandler(ErrorCategory category, ref string message)
		{
			if (category == ErrorCategory.OperationCanceled)
			{
				message = "Index opening was canceled";
				return true;
			}
			return false;
		}
		
		private bool DefaultErrorHandler(ErrorCategory category, ref string message)
		{
			if (category == ErrorCategory.OperationCanceled)
			{
				// ignore it
				message = null;
				return true;
			}
			return false;
		}
		
		private bool SaveIndexErrorHandler(ErrorCategory category, ref string message)
		{
			if (category == ErrorCategory.OperationCanceled)
			{
				message = "Index saving was canceled";
				return true;
			}
			return false;
		}
		
		private void InternalAddFiles(string[] fileNames)
		{
			Write("Update index");
			LockControls(InterruptMode.Enabled);
			try
			{
				try
				{
					ClearAllResults();
					AddToIndexCommand command = new AddToIndexCommand();
					command.Index = _index;
					command.FileNames = fileNames;
					_worker.Execute(command);
				}
				catch(Exception error)
				{
					ShowException(error, this.UpdateIndexErrorHandler);
				}
			}
			finally
			{
				WriteLastError();
				UnLockControls(FocusMode.Keep);
				UpdateFileCount();
			}
		}
		
		void MenuIndexAddFileClick(object sender, EventArgs e)
		{
			if (addFileDialog.ShowDialog() == DialogResult.OK)
			{
				InternalAddFiles(addFileDialog.FileNames);
			}
		}
		
		void MenuIndexAddFolderClick(object sender, EventArgs e)
		{
			if (addFolderDialog.ShowDialog() == DialogResult.OK)
			{
				InternalAddFiles(new string[] {addFolderDialog.SelectedPath});
			}
		}
		
		void MenuIndexAddMountClick(object sender, EventArgs e)
		{
			if (addMountDialog.ShowDialog() == DialogResult.OK)
			{
				InternalAddMounts(addMountDialog.FileNames);
			}
		}
		
		private const int HistorySize = 100; // TUNE:
		
		private void AddItemToHistory(string subject, System.Windows.Forms.ToolStripComboBox list, string text)
		{
			string oldText = list.Text;
			try
			{
				if (String.IsNullOrEmpty(text))
				{
					return;
				}
				
				for(int i = 0; i < list.Items.Count; i++)
				{
					string item = (string)list.Items[i];
					if (item == text)
					{
						list.Items.RemoveAt(i);
						break;
					}
				}
				
				list.Items.Insert(1 /* we are skipping the fixed 0-th blank entry */, text);
				while (list.Items.Count >= HistorySize)
				{
					list.Items.RemoveAt(list.Items.Count - 1);
				}
			}
			finally
			{
				list.Text = oldText;
			}
			
			if (!String.IsNullOrEmpty(subject))
			{
				_synchronizer.Broadcast(subject, text);
			}
		}

		private int SecondsFrom(DateTime start)
		{
			TimeSpan diff = DateTime.Now.Subtract(start).Duration();
			int seconds = diff.Seconds;
			if (diff.Milliseconds >= 500 /* TODO: is this the right rounding limit? */)
			{
				seconds++;
			}
			return seconds + 60 * (diff.Minutes + 60 * diff.Hours);
		}
		
		private void SetSearchStatus(int count)
		{
			UpdateFoundDocumentCount(count);
			this.statusActivity.Text = "";
		}

		private enum HistoryDirection
		{
			DownUp,
			UpDown
		}
		
		private void ClearHistoryItems(System.Windows.Forms.ToolStripSplitButton owner)
		{
			owner.DropDownItems.Clear();
		}
		
		private void EnableHistoryButton(System.Windows.Forms.ToolStripSplitButton owner)
		{
			if (owner.DropDownItems.Count > 0)
			{
				owner.Enabled = true;
				owner.Text = owner.DropDownItems[0].Text;
			}
			else
			{
				owner.Enabled = false;
				owner.Text = "";
			}
		}

		private void HistoryItemClick(object sender, EventArgs e)
		{
			ActivateSearch((Search)(((ToolStripMenuItem)sender).Tag));
		}
		
		private void AddHistoryItem(System.Windows.Forms.ToolStripSplitButton owner, Search item, HistoryDirection direction)
		{
			System.Windows.Forms.ToolStripMenuItem menu = new ToolStripMenuItem();
			menu.Text = item.DisplayLabel;
			menu.Tag = item;
			menu.Click += this.HistoryItemClick;
			switch (direction)
			{
				case HistoryDirection.DownUp:
					owner.DropDownItems.Insert(0, menu);
					break;
					
				case HistoryDirection.UpDown:
					owner.DropDownItems.Add(menu);
					break;
					
				default:
					throw new NotImplementedException();
			}
		}
		
		private void UpdateSearchHistoryButtons()
		{
			int currentIndex = _history.IndexOf(_search);
			if (currentIndex < 0)
			{
				currentIndex = _history.Count;
			}
			
			ClearHistoryItems(this.goBackButton);
			ClearHistoryItems(this.goForwardButton);
			
			for(int i = 0; i < currentIndex; i++)
			{
				AddHistoryItem(this.goBackButton, _history[i], HistoryDirection.DownUp);
			}
			
			for(int i = currentIndex + 1; i < _history.Count; i++)
			{
				AddHistoryItem(this.goForwardButton, _history[i], HistoryDirection.UpDown);
			}
			
			EnableHistoryButton(this.goBackButton);
			EnableHistoryButton(this.goForwardButton);
		}

		private void DisconnectSearch()
		{
			if (_search != null)
			{
				_search.SetSelectionFromListView(this.fileList);
			}
			_search = null;			
		}
		
		private void ClearResults(bool preserveSearchMode)
		{
			DisconnectSearch();
			this.fileList.VirtualListSize = 0;
			this.fileList.VirtualMode = false;
			this.fileList.Items.Clear();
			this.fileList.Invalidate();
			this.fileList.Update();
			_itemCache = null;
			UpdateSearchHistoryButtons();
			if (!preserveSearchMode)
			{
				SearchMode = Index.SetCombineMode.None;
			}
			UpdateFoundDocumentCount(0);
			UpdateFileCount();
		}
		
		private void ClearAllResults()
		{
			_history.Clear();
			_search = null;
			ClearResults(false);
			UpdateSearchHistoryButtons();
		}
		
		private bool _resizing = false;
		private void ResizeFileList()
		{
			if (!_resizing)
			{
				_resizing = true;
				
				// TODO: move the code to the ListViewHelper
				
				// first get size estimate without the possible vertical scrollbar
				this.columnFileName.Width = this.fileList.ClientSize.Width - /* TODO: get the number from somewhere */ 20;

				// then call it again while the horizontal scrollbar is calculated-away
				this.columnFileName.Width = this.fileList.ClientSize.Width - /* TODO: get the number from somewhere */ 2;

				// Fix ListView bug
				if ((!ListViewHelper.VerticalScrollBarVisible(this.fileList)) && (this.fileList.Items.Count > 0))
				{
					/*
					this.fileList.Scrollable = false;
					this.fileList.Update();
					this.fileList.Scrollable = true;
					this.fileList.Update();
					*/
					/* TODO: ??
					this.fileList.Scrollable = false;
					this.fileList.Scrollable = true;
					// TODO: listview is vertically offset by a non-zero. How can we fix it?
					ListViewHelper.ResetTopItem(this.fileList);
					*/
					ListViewHelper.ResetTopItem(this.fileList);
				}
				
				_resizing = false;
			}
		}

		private void ProcessSyncMessage(System.Threading.MessageBoard.Entry message)
		{
			switch (message.Subject)
			{
				case "SearchHistory":
					AddItemToHistory(null, this.searchBox, message.Body);
					break;
					
				case "FileMaskHistory":
					AddItemToHistory(null, this.fileMaskList, message.Body);
					break;
			}
		}

		void InternalSearch(Index.SetCombineMode searchMode, string text, string mask, bool caseSensitive, bool wholeWords)
		{
			int count = 0;
			
			Write("Search");
			LockControls(InterruptMode.Enabled);
			try
			{
				try
				{
					this.searchBox.Text = text;
					this.fileMaskList.Text = mask;
					this.caseInSensitiveCheckBox.Checked = !caseSensitive;
					this.partialWordCheckBox.Checked = !wholeWords;
					this.SearchMode = searchMode;
					
					Search lastSearch = _search;
					Index.Query lastQuery = _search == null ? null : _search.Query;
					
					this.statusActivity.Text = "";
					AddItemToHistory("SearchHistory", this.searchBox, this.searchBox.Text);
					AddItemToHistory("FileMaskHistory", this.fileMaskList, this.fileMaskList.Text);
					
					Index.Query query = Index.Query.New(lastQuery, SearchMode, this.searchBox.Text, !this.caseInSensitiveCheckBox.Checked, !this.partialWordCheckBox.Checked, new System.IO.FileFilter(this.fileMaskList.Text));
					
					Index.DocumentCollection list = null;
					
					SearchCommand command = new SearchCommand();
					command.Index = _index;
					command.Query = query;
	
					ClearResults(true);
					
					_worker.Execute(command);
					
					list = command.Result;
	
					count = list.Count;
					SetSearchStatus(count);
					
					Search results = new Search(query, this.fileMaskList.Text, list);
					results.CaseSensitive = !this.caseInSensitiveCheckBox.Checked;
					results.WholeWords = !this.partialWordCheckBox.Checked;
					results.Phrase = this.searchBox.Text;
					_history.InsertAfter(results, lastSearch);
					_search = results;
	
					UpdateSearchHistoryButtons();
					
					this.fileList.VirtualListSize = list.Count;
					this.fileList.VirtualMode = true;
					ResizeFileList();
					
					if (_interrupted)
					{
						throw new OperationCanceledException();
					}
				}
				catch(Exception error)
				{
					ShowException(error, this.SearchErrorHandler);
				}
			}
			finally
			{
				WriteLastError();
				SetSearchStatus(count);
				UnLockControls(FocusMode.Auto);
			}
		}
		
		void SearchButtonClick(object sender, EventArgs e)
		{
			InternalSearch(SearchMode, this.searchBox.Text, this.fileMaskList.Text, !this.caseInSensitiveCheckBox.Checked, !this.partialWordCheckBox.Checked);
		}
		
		private void ActivateSearch(Search item)
		{
			ClearResults(true);
			
			this.statusTimer.Text = T2S(0);
			
			_search = item;
			
			this.caseInSensitiveCheckBox.Checked = !_search.CaseSensitive;
			this.partialWordCheckBox.Checked = !_search.WholeWords;
			this.searchBox.Text = _search.Phrase;
			this.fileMaskList.Text = _search.Masks;
			
			UpdateSearchHistoryButtons();
			this.fileList.VirtualListSize = _search.Results.Count;
			this.fileList.VirtualMode = true;
			ResizeFileList();
			
			if (_search != null)
			{
				_search.SetSelectionToListView(this.fileList);
			}
			
			SetSearchStatus(_search.Results.Count);
		}

		private enum ErrorCategory
		{
			Other,
			OperationCanceled
		}
		
		private delegate bool TranslateError(ErrorCategory category, ref string message);
		
		private void ShowException(Exception error, TranslateError handler)
		{
			ErrorCategory category = ErrorCategory.Other;
			if (error is OperationCanceledException)
			{
				category = ErrorCategory.OperationCanceled;
			}
			string message = error.Message;
			if (handler != null)
			{
				if (handler(category, ref message))
				{
					_lastError = message;
					return;
				}
			}
			MessageBox.Show(message, this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
		}
		
		void MenuFileOpenClick(object sender, EventArgs e)
		{
			try
			{
				try
				{
					if (openIndexDialog.ShowDialog() == DialogResult.OK)
					{
						if (InternalCloseIndex())
						{
							try
							{
								InternalOpenIndex(openIndexDialog.FileName);
							}
							finally
							{
								UpdateIndexControls();
							}
						}
					}
				}
				catch (Exception error)
				{
					ShowException(error, this.OpenIndexErrorHandler);
				}
			}
			finally
			{
				WriteLastError();
			}
		}
		
		void MenuFileDropDownOpening(object sender, EventArgs e)
		{
			this.menuFileSave.Enabled = (_index != null) && _index.Modified;
			this.menuFileSaveAs.Enabled = (_index != null);
			this.menuFileClose.Enabled = (_index != null);
		}
		
		void MenuIndexDropDownOpening(object sender, EventArgs e)
		{
			bool indexValid = _index != null;
			bool canMountArchive = false;
			bool canAddFile = false;
			
			if (indexValid)
			{
				if (_index.Mounts.Length > 0)
				{
					// some archive is already mounted
					canMountArchive = true;
					canAddFile = false;
				}
				else
				{
					if (_index.Seeds.Length > 0)
					{
						// some file was already added
						canMountArchive = false;
						canAddFile = true;
					}
					else
					{
						// virgin
						canMountArchive = true;
						canAddFile = true;
					}
				}
			}
			
			this.menuIndexAddMount.Enabled = canMountArchive;
			this.menuIndexAddFile.Enabled = canAddFile;
			this.menuIndexAddFolder.Enabled = canAddFile;
			this.menuIndexRebuild.Enabled = indexValid;
			this.menuIndexUpdate.Enabled = indexValid;
			this.menuIndexProperties.Enabled = indexValid;
		}

		void FileListDoubleClick(object sender, EventArgs e)
		{
			ValidateFileListMenu();
			foreach(ToolStripItem item in this.fileListContextMenu.Items)
			{
				if (IsDefault(item))
				{
					item.PerformClick();
					return;
				}
			}
		}

		private const int MaxNumberOfFilesToPutToClipboard = 1000;	// TODO: TUNE:
		
		void FileListMenuCopyClick(object sender, EventArgs e)
		{
			Wait(true);
			try
			{
				System.Text.StringBuilder buffer = new System.Text.StringBuilder();
				string[] fileNames = null;
				
				bool copyFiles = _index.UsesNativeIO && (fileList.SelectedIndices.Count <= MaxNumberOfFilesToPutToClipboard);

				if (copyFiles)
				{
					fileNames = new string[fileList.SelectedIndices.Count];
				}
				
				int indexToNames = 0;
				foreach(int index in fileList.SelectedIndices)
				{
					string fileName = _search.Results[index];
					buffer.AppendLine(fileName);
					
					if (copyFiles && ((_search.Results.GetItemKind(index) & Index.DocumentKind.Deleted) == 0))
					{
						fileNames[indexToNames] = fileName;
						indexToNames++;
					}
				}
				DataObject fileContainer = new DataObject();
				if (copyFiles)
				{
					Array.Resize(ref fileNames, indexToNames);
					fileContainer.SetData(DataFormats.FileDrop, true, fileNames);
				}
				fileContainer.SetText(buffer.ToString());
				System.Windows.Forms.Clipboard.SetDataObject(fileContainer, true);
			}
			finally
			{
				Wait(false);
			}
		}
		
		void OpenCurrentFile(bool ask)
		{
			try
			{
				try
				{
					LockControls(InterruptMode.Disabled);
					try
					{
						OpenFileCommand opener = new OpenFileCommand();
						opener.FileName = _search.Results[this.fileList.SelectedIndices[0]];
						opener.IO = _index.IO;
						opener.TemporaryFiles = _temporaryFiles;
						if (ask)
						{
							opener.Mode = OpenFileCommand.OpenMode.Ask;
						}
						else
						{
							opener.Mode = OpenFileCommand.OpenMode.OpenOrAsk;
						}
						_worker.Execute(opener);
					}
					finally
					{
						UnLockControls(FocusMode.Keep);
					}
				}
				catch (Exception error)
				{
					ShowException(error, this.DefaultErrorHandler);
				}
			}
			finally
			{
				WriteLastError();
			}
		}
		
		void FileListMenuOpenClick(object sender, EventArgs e)
		{
			OpenCurrentFile(false);
		}
		
		void FileListMenuOpenWithClick(object sender, EventArgs e)
		{
			OpenCurrentFile(true);
		}
		
		void FileListMenuOpenFolderClick(object sender, EventArgs e)
		{
			try
			{
				try
				{
					LockControls(InterruptMode.Disabled);
					try
					{
						OpenContainingFolderCommand opener = new OpenContainingFolderCommand();
						opener.FileNames = new string[this.fileList.SelectedIndices.Count];
						for(int i = 0; i < opener.FileNames.Length; i++)
						{
							opener.FileNames[i] = _search.Results[this.fileList.SelectedIndices[i]];
						}
						opener.IO = _index.IO;
						opener.TemporaryFiles = _temporaryFiles;
						_worker.Execute(opener);
					}
					finally
					{
						UnLockControls(FocusMode.Keep);
					}
				}
				catch (Exception error)
				{
					ShowException(error, this.DefaultErrorHandler);
				}
			}
			finally
			{
				WriteLastError();
			}
		}
		
		void MainFormFormClosing(object sender, FormClosingEventArgs e)
		{
			if (Locked)
			{
				e.Cancel = true;
				return;
			}

			_synchronizer.ReceiveAll();
			
			ISettingsWriter config = GetConfig();
			try
			{
				SaveLayout(config);
			}
			finally
			{
				config.Dispose();
			}
			
			if (!InternalCloseIndex())
			{
				e.Cancel = true;
			}
			
			if (!e.Cancel)
			{
				foreach(string fileName in _temporaryFiles)
				{
					try
					{
						if (System.IO.Directory.Exists(fileName))
						{
							foreach(string item in System.IO.Directory.GetFiles(fileName, "*.*", System.IO.SearchOption.AllDirectories))
							{
								// clear read-only marks
								System.IO.File.SetAttributes(item, System.IO.FileAttributes.Normal);
							}
							System.IO.Directory.Delete(fileName, true);
						}
						else
						{
							System.IO.File.Delete(fileName);
						}
					}
					catch (Exception)
					{
					}
				}
			}
		}
		
		void MenuFileCloseClick(object sender, EventArgs e)
		{
			if (InternalCloseIndex())
			{
				SetLastIndex(null);
				UpdateIndexControls();
			}
		}
		
		private string GetItemFolder(int itemIndex)
		{
			return System.IO.Path.GetDirectoryName(_search.Results[itemIndex]);
		}
		
		private bool IsValidTextFile(int selectionIndex)
		{
			Index.DocumentKind kind = _search.Results.GetItemKind(this.fileList.SelectedIndices[selectionIndex]);
			
			if ((kind & Index.DocumentKind.Deleted) != 0)
			{
				return false;
			}
			
			switch (kind)
			{
				case Index.DocumentKind.Indexed:
					return true;

				case Index.DocumentKind.Indexed | Index.DocumentKind.Cached:
					// original file is probably binary or otherwise non-comparable
					return false;
					
				default:
					return false;
			}
		}
		
		void FileListContextMenuOpening(object sender, System.ComponentModel.CancelEventArgs e)
		{
			bool selectionValid = this.fileList.SelectedIndices.Count > 0;
			bool singleFile = this.fileList.SelectedIndices.Count == 1;
			bool validSingleFile = singleFile;
			bool singleFolder = selectionValid;
			
			if (singleFolder && !singleFile)
			{
				// verify that all selected files reside in the same folder
				string folder = GetItemFolder(this.fileList.SelectedIndices[0]);
				for(int i = 1; i < this.fileList.SelectedIndices.Count; i++)
				{
					if (String.Compare(folder, GetItemFolder(this.fileList.SelectedIndices[i]), true) != 0)
					{
						singleFolder = false;
						break;
					}
				}
			}
			
			bool useShowAsDefault = true;
			
			if (singleFile)
			{
				Index.DocumentKind kind = _search.Results.GetItemKind(this.fileList.SelectedIndices[0]);
				
				if ((kind & Index.DocumentKind.Deleted) != 0)
				{
					validSingleFile = false;
				}
				else if ((kind & Index.DocumentKind.Indexed) == 0)
				{
					// for not-indexed files the "show" is not the default
					useShowAsDefault = false;
				}
			}
			
			this.fileListMenuCopy.Enabled = selectionValid;
			this.fileListMenuShow.Enabled = selectionValid;
			this.fileListMenuOpen.Enabled = validSingleFile;
			this.fileListMenuOpenWith.Enabled = validSingleFile;
			this.fileListMenuOpenFolder.Enabled = singleFolder;
			this.fileListMenuHide.Enabled = false; // TODO: selectionValid;
			this.fileListMenuFindReferences.Enabled = selectionValid;
			this.fileListMenuSelectAll.Enabled = this.fileList.Items.Count > 0;
			this.fileListMenuProperties.Enabled = selectionValid;
			this.fileListMenuCompare.Enabled = 
				selectionValid &&
				(this.fileList.SelectedIndices.Count == 2) &&
				IsValidTextFile(0) &&
				IsValidTextFile(1) &&
				GetComparer();
			
			SetDefault(this.fileListMenuShow, useShowAsDefault);
			SetDefault(this.fileListMenuOpen, !useShowAsDefault);
		}
		
		private bool IsDefault(ToolStripItem item)
		{
			return item.Font.Bold;
		}
		
		private void SetDefault(ToolStripMenuItem item, bool value)
		{
			// all we want to achieve with bolding and unbolding is some visual marker indicating "default" item
			if (IsDefault(item) == value)
			{
				return;
			}
			System.Drawing.FontStyle style = value ? System.Drawing.FontStyle.Bold : System.Drawing.FontStyle.Regular;
			item.Font = new System.Drawing.Font(item.Font, style);
		}
		
		void MainFormLoad(object sender, EventArgs e)
		{
			Write("Initialize");
			IdleTask.New(this.StartUp);
		}

		private void LoadComboBoxLayout(ISettingsReader config, string section, System.Windows.Forms.ToolStripComboBox list)
		{
			list.Items.Clear();
			list.Items.Add("");
			for(int i = 0; true; i++)
			{
				string item = config.ReadString(String.Format("{0}/{1}", section, i), null);
				if (item == null)
				{
					break;
				}
				if (!String.IsNullOrEmpty(item))
				{
					list.Items.Add(item);
				}
			}
			list.Text = config.ReadString(String.Format("{0}/Selection", section), null);
		}
		
		private void LoadLayout(ISettingsReader config)
		{
			this.partialWordCheckBox.Checked = !config.ReadBoolean("Search/WholeWord", !this.partialWordCheckBox.Checked);
			this.caseInSensitiveCheckBox.Checked = !config.ReadBoolean("Search/CaseSensitive", !this.caseInSensitiveCheckBox.Checked);

			System.Windows.Forms.FormLayout layout = new System.Windows.Forms.FormLayout();
			layout.Form = this;
			layout.Load(config, "Window");
			
			LoadComboBoxLayout(config, "SearchHistory", this.searchBox);
			LoadComboBoxLayout(config, "FileMaskHistory", this.fileMaskList);
		}

		private void SaveComboBoxLayout(ISettingsWriter config, string section, System.Windows.Forms.ToolStripComboBox list)
		{
			config.Delete(String.Format("{0}/", section));
			int id = 0;
			for(int i = 0; i < list.Items.Count; i++)
			{
				string item = (string)list.Items[i];
				if (!String.IsNullOrEmpty(item))
				{
					config.WriteString(String.Format("{0}/{1}", section, id), item);
					id++;
				}
			}
			string last = list.Text;
			if (!String.IsNullOrEmpty(last))
			{
				config.WriteString(String.Format("{0}/Selection", section), last);
			}
		}
		
		private void SaveLayout(ISettingsWriter config)
		{
			config.WriteBoolean("Search/WholeWord", !this.partialWordCheckBox.Checked);
			config.WriteBoolean("Search/CaseSensitive", !this.caseInSensitiveCheckBox.Checked);
			
			System.Windows.Forms.FormLayout layout = new System.Windows.Forms.FormLayout();
			layout.Form = this;
			layout.Save(config, "Window");
			
			SaveComboBoxLayout(config, "SearchHistory", this.searchBox);
			SaveComboBoxLayout(config, "FileMaskHistory", this.fileMaskList);
		}
		
		private void StartUp()
		{
			try
			{
				ISettingsReader config = GetConfig();
				try
				{
					LoadLayout(config);
					
					string fileName;
					if (System.Configuration.CommandLine.Args.Length == 1)
					{
						fileName = System.Configuration.CommandLine.Args[0];
					}
					else
					{
						fileName = config.ReadString("Path/LastIndex", null);
					}
					try
					{
						InternalOpenIndex(fileName);
					}
					catch(Exception)
					{
						InternalOpenIndex(null);
					}
				}
				finally
				{
					config.Dispose();
				}
				_synchronizer = new System.Windows.Forms.MessageBoardClient(new Guid("{DDE2250F-9421-4A3D-BCE4-6F91BF8B2669}"), this.ProcessSyncMessage);
			}
			finally
			{
				Write();
			}
		}
		
		void MenuIndexRebuildClick(object sender, EventArgs e)
		{
			Write("Rebuild index");
			LockControls(InterruptMode.Enabled);
			try
			{
				try
				{
					ClearAllResults();
					RebuildIndexCommand command = new RebuildIndexCommand();
					command.Index = _index;
					_worker.Execute(command);
				}
				catch(Exception error)
				{
					ShowException(error, this.UpdateIndexErrorHandler);
				}
			}
			finally
			{
				WriteLastError();
				UnLockControls(FocusMode.Keep);
				UpdateFileCount();
			}
		}

		protected static bool KeyPressed(Keys key) 
		{ 
			return (Control.ModifierKeys & key) == key; 
		} 
		
		void MenuIndexUpdateClick(object sender, EventArgs e)
		{
			Write("Update index");
			LockControls(InterruptMode.Enabled);
			try
			{
				try
				{
					ClearAllResults();
					UpdateIndexCommand command = new UpdateIndexCommand();
					command.SkipCheckingPhase = KeyPressed(Keys.Shift);
					command.Index = _index;
					_worker.Execute(command);
				}
				catch(Exception error)
				{
					ShowException(error, this.UpdateIndexErrorHandler);
				}
			}
			finally
			{
				WriteLastError();
				UnLockControls(FocusMode.Keep);
				UpdateFileCount();
			}
		}
		
		
		#region IIndexObserver interface implementation
		public void DocumentCountChanged(int documentCount)
		{
			UpdateFileCount(documentCount);
		}
		
		public void FoundDocumentCountChanged(int documentCount)
		{
			UpdateFoundDocumentCount(documentCount);
		}
		
		public void OpenProgress(string message, int count)
		{
			_escape = false;
			this.statusProgress.Minimum = 0;
			this.statusProgress.Value = 0;
			this.statusProgress.Maximum = count;
			this.statusProgress.Style = ProgressBarStyle.Continuous;
			this.statusProgress.Visible = true;
			WriteProgress(message);
		}

		public void ShowMessage(string message)
		{
			WriteProgress(message);
		}
		
		private bool _escape = false;
		private bool _interrupted = false;
		public bool ShowProgress(int index)
		{
			this.statusProgress.Value = index;
			return !InterruptRequested();
		}

		public bool InterruptRequested()
		{
			if (_escape)
			{
				if (!_interrupted)
				{
					// show "interrupted" indicator
					this.statusError.Text = "";
					this.statusError.Visible = true;
					_interrupted = true;
				}
				return true;
			}
			else
			{
				return false;
			}
		}
		
		public void CloseProgress()
		{
			WriteProgress();
			this.statusProgress.Visible = false;
		}
		#endregion
		
		bool DoSave()
		{
			if (String.IsNullOrEmpty(_index.FileName))
			{
				return DoSaveAs();
			}
			else
			{
				LockControls(InterruptMode.Enabled);
				Write("Save");
				try
				{
					SaveIndexCommand command = new SaveIndexCommand();
					command.Index = _index;
					try
					{
						_worker.Execute(command);
					}
					catch(Exception error)
					{
						ShowException(error, this.SaveIndexErrorHandler);
						return false;
					}
				}
				finally
				{
					WriteLastError();
					UnLockControls(FocusMode.Keep);
				}
				return true;
			}
		}

		bool DoSaveAs()
		{
			this.saveIndexDialog.FileName = this.openIndexDialog.FileName;
			if (this.saveIndexDialog.ShowDialog() == DialogResult.OK)
			{
				LockControls(InterruptMode.Enabled);
				Write("Save");
				try
				{
					SaveIndexCommand command = new SaveIndexCommand();
					command.Index = _index;
					command.FileName = this.saveIndexDialog.FileName;
					try
					{
						_worker.Execute(command);
					}
					catch(Exception error)
					{
						ShowException(error, this.SaveIndexErrorHandler);
						return false;
					}
					SetLastIndex(_index.FileName);
					this.openIndexDialog.FileName = _index.FileName;
					UpdateIndexControls();
				}
				finally
				{
					WriteLastError();
					UnLockControls(FocusMode.Keep);
				}
				return true;
			}
			else
			{
				return false;
			}
		}
		
		void MenuFileSaveClick(object sender, EventArgs e)
		{
			DoSave();
		}
		
		void MenuFileSaveAsClick(object sender, EventArgs e)
		{
			DoSaveAs();
		}
		
		void FileListMenuHideClick(object sender, EventArgs e)
		{
			if (fileList.SelectedIndices.Count > 0)
			{
				int[] indices = new int[fileList.SelectedIndices.Count];
				fileList.SelectedIndices.CopyTo(indices, 0);
				for(int i = 0; i < indices.Length; i++)
				{
					int index = indices[i];
					fileList.Items.RemoveAt(index);
					for(int j = 0; j < indices.Length; j++)
					{
						if (indices[j] > index)
						{
							indices[j] = indices[j] - 1;
						}
					}
				}
			}
		}
		
		void SearchBoxKeyDownUp(object sender, KeyEventArgs e)
		{
			if ((e.KeyCode == Keys.Enter) && !this.searchBox.DroppedDown)
			{
				e.Handled = true;
				e.SuppressKeyPress = true;
				this.searchButton.PerformClick();
			}
		}

		private void ValidateFileListMenu()
		{
			this.FileListContextMenuOpening(null, new System.ComponentModel.CancelEventArgs());
		}
		
		private void ValidateIndexMenu()
		{
			this.MenuIndexDropDownOpening(null, new System.EventArgs());
		}
		
		void FileListKeyDownUp(object sender, KeyEventArgs e)
		{
			/* TODO: hiding not supported in virtual mode
			if (e.KeyCode == Keys.Delete)
			{
				e.Handled = true;
				e.SuppressKeyPress = true;
				this.fileListMenuHide.PerformClick();
			}
			*/
			if (e.KeyCode == Keys.Enter)
			{
				e.Handled = true;
				FileListDoubleClick(null, null);
			}
			else if ((e.KeyCode == Keys.A) && e.Control)
			{
				if (this.fileList.Enabled)
				{
					ListViewHelper.SelectAll(this.fileList, true);
					e.Handled = true;
				}
			}
			else if (((e.KeyCode == Keys.C) && e.Control) || ((e.KeyCode == Keys.Insert) && e.Control))
			{
				e.Handled = true;
				ValidateFileListMenu();
				this.fileListMenuCopy.PerformClick();
			}
		}
		
		void MainFormPreviewKeyDown(object sender, PreviewKeyDownEventArgs e)
		{
			// TODO: why this doesn't work?
			if ((e.KeyCode == Keys.F) && (e.Control))
			{
				this.searchBox.Focus();
			}
		}
		
		void MenuHelpAboutClick(object sender, EventArgs e)
		{
			MessageForm message = new MessageForm(this);
			message.Text = "About Indexer";
			
			System.Version version = System.Reflection.Assembly.GetExecutingAssembly().GetName().Version;
			
			message.AddGroup(String.Format("Indexer version {0}.{1}.{2}", version.Major, version.Minor, version.Build));
			message.AddValue("Brought to you by Mojmr Nmeek, (http://xmojmr.ohmygod.cz/software/Indexer)");
			
			// TODO: make sure the message covers all 3rd party licensing issues
			message.AddGroup("This product includes following 3rd party components");
			message.AddValue("Apache Lucene.Net (http://incubator.apache.org/lucene.net)");
			message.AddValue("ICSharpCode.net ZipLibrary (http://sharpdevelop.net/OpenSource/SharpZipLib)");
			message.AddValue("Colorer Library syntax coloring/regexp codes by Igor Russkih, Cail Lomecb (http://colorer.sourceforge.net)");
			message.AddValue("Using IFilter in C# (http://www.codeproject.com/KB/cs/IFilter.aspx)");
			
			message.Show();
		}
		
		void MainFormKeyDown(object sender, KeyEventArgs e)
		{
			if ((e.KeyCode == Keys.Escape) && Locked)
			{
				e.Handled = true;
				if (this.interruptButton.Enabled)
				{
					_escape = true;
				}
			}
			else if (!Locked)
			{
				if ((e.KeyCode == Keys.F) && e.Control)
				{
					e.Handled = true;
					this.searchBox.Focus();
				}
			}
		}
		
		private string N2S(int value)
		{
			return value.ToString("#,###,###,##0");
		}
		
		private string T2S(int seconds)
		{
			int minutes = seconds / 60;
			seconds = seconds % 60;
			int hours = minutes / 60;
			minutes = minutes % 60;
			return hours.ToString("#0") + ":" + minutes.ToString("00") + ":" + seconds.ToString("00");
		}
		
		private string T2S(DateTime time)
		{
			return time.ToLongDateString() + " " + time.ToShortTimeString();
		}
		
		private string N2S(long value)
		{
			return value.ToString("#,###,###,##0");
		}
		
		private string N2S(string format, long value)
		{
			string suffix;
			if (value == 1)
			{
				suffix = "";
			}
			else
			{
				suffix = "s";
			}
			return String.Format(format, N2S(value), suffix);
		}

		private string N2S(string format, int value)
		{
			return N2S(format, (long)value);
		}
		
		void MenuIndexPropertiesClick(object sender, EventArgs e)
		{
			MessageForm message = new MessageForm(this);
			message.Text = "Index Properties";
			
			if (!String.IsNullOrEmpty(_index.FileName))
			{
				message.Add("File", String.Format("\"{0}\"", _index.FileName));
			}
			
			DateTime lastWriteTime = _index.LastWriteTime;
			message.Add("Updated", T2S(lastWriteTime));
			
			message.AddGroup("Archives");
			foreach(string archive in _index.Mounts)
			{
				message.AddValue(String.Format("\"{0}\"", archive));
			}
			
			message.AddGroup("Seeds");
			foreach(string seed in _index.Seeds)
			{
				message.AddValue(String.Format("\"{0}\"", seed));
			}
			
			message.AddGroup("Metrics");
			message.AddValue(N2S("{0} unpacked byte{1}", _index.Size));
			message.AddValue(N2S("{0} document{1}", _index.DocumentsCount));
			long documentsSize = _index.DocumentsSize;
			if (documentsSize >= 0)
			{
				message.AddValue(N2S("{0} document byte{1}", documentsSize));
			}
			message.AddValue(N2S("{0} word{1}", _index.WordCount));
			message.AddValue(N2S("{0} letter{1}", _index.WordLetters));
			
			message.Show();
		}
		
		private class ListViewItemCache
		{
			private System.Collections.Generic.List<ListViewItem> _items;
			private int _startIndex;
			private int _endIndex;
			public ListViewItemCache(int startIndex, int endIndex)
			{
				_startIndex = startIndex;
				_endIndex = endIndex;
				_items = new System.Collections.Generic.List<ListViewItem>(_endIndex - _startIndex + 1);
			}
			
			public ListViewItem this[int index]
			{
				get
				{
					if ((index < _startIndex) || (index > _endIndex))
					{
						return null;
					}
					else
					{
						index -= _startIndex;
						if (index < _items.Count)
						{
							return _items[index];
						}
						else
						{
							return null;
						}
					}
				}
				set
				{
					if ((index < _startIndex) || (index > _endIndex))
					{
						return;
					}
					index -= _startIndex;
					if (index < _items.Count)
					{
						_items[index] = value;
					}
					else
					{
						for(int i = _items.Count; i < index; i++)
						{
							_items.Add(null);
						}
						_items.Add(value);
					}
				}
			}
		}
		private ListViewItemCache _itemCache = null;

		private ListViewItem CreateListViewItem(int index)
		{
			ListViewItem result = new ListViewItem(_search.Results[index]);
			Index.DocumentKind kind = _search.Results.GetItemKind(index);
			if (kind == Index.DocumentKind.Indexed)
			{
				result.ImageIndex = 0;
			}
			else if (kind == (Index.DocumentKind.Indexed | Index.DocumentKind.Cached))
			{
				result.ImageIndex = 2;
			}
			else if ((kind & Index.DocumentKind.Deleted) != 0)
			{
				result.ForeColor = System.Drawing.SystemColors.GrayText;
				result.ImageIndex = -1;
			}
			else if ((kind == Index.DocumentKind.Binary) || (kind == (Index.DocumentKind.Binary | Index.DocumentKind.Cached)))
			{
				result.ImageIndex = 1;
			}
			else
			{
				throw new NotImplementedException();
			}
			return result;
		}
		
		void FileListCacheVirtualItems(object sender, CacheVirtualItemsEventArgs e)
		{
			/* TODO: this immediatelly loading code doesn't seem to improve performance
			Cursor oldCursor = this.Cursor;
			this.Cursor = Cursors.WaitCursor;
			try
			{
				_itemCache = new ListViewItemCache(e.StartIndex, e.EndIndex);
				for(int i = e.StartIndex; i <= e.EndIndex; i++)
				{
					_itemCache[i] = CreateListViewItem(i);
				}
			}
			finally
			{
				this.Cursor = oldCursor;
			}
			*/
			_itemCache = new ListViewItemCache(e.StartIndex, e.EndIndex);
		}
		
		void FileListRetrieveVirtualItem(object sender, RetrieveVirtualItemEventArgs e)
		{
			if (_itemCache != null)
			{
				e.Item = _itemCache[e.ItemIndex];
			}
			else
			{
				e.Item = null;
			}
			
			if (e.Item == null)
			{
				e.Item = CreateListViewItem(e.ItemIndex);
				if (_itemCache != null)
				{
					_itemCache[e.ItemIndex] = e.Item;
				}
			}
		}
		
		void FileListResize(object sender, EventArgs e)
		{
			ResizeFileList();
		}
		
		void FileListColumnClick(object sender, ColumnClickEventArgs e)
		{
			//
		}
		
		void MainFormFormClosed(object sender, FormClosedEventArgs e)
		{
			if (_worker != null)
			{
				_worker.Dispose();
				_worker = null;
			}
			if (_synchronizer != null)
			{
				_synchronizer.Dispose();
				_synchronizer = null;
			}
		}
		
		void SetOperatorListSelectedIndexChanged(object sender, EventArgs e)
		{
			string selection = this.setOperatorList.Text;
			if (String.IsNullOrEmpty(selection))
			{
				selection = "";
			}
			
			foreach(Index.SetCombineMode mode in _searchModeMap.Keys)
			{
				SearchModeEntry entry = _searchModeMap[mode];
				if (entry.ListValue == selection)
				{
					SearchMode = mode;
					break;
				}
			}
		}
		
		void CaseInSensitiveCheckBoxClick(object sender, EventArgs e)
		{
			this.caseInSensitiveCheckBox.Checked = !this.caseInSensitiveCheckBox.Checked;
		}
		
		void PartialWordCheckBoxClick(object sender, EventArgs e)
		{
			this.partialWordCheckBox.Checked = !this.partialWordCheckBox.Checked;
		}
		
		void ClearSearchResultsClick(object sender, EventArgs e)
		{
			ClearResults(false);
		}
		
		void FileListMenuShowClick(object sender, EventArgs e)
		{
			OpenSelectedFiles();
		}
		
		void InterruptButtonClick(object sender, EventArgs e)
		{
			if (Locked)
			{
				if (this.interruptButton.Enabled)
				{
					_escape = true;
				}
			}
		}
		
		void BackgroundModeClick(object sender, EventArgs e)
		{
			this.backgroundMode.Checked = !this.backgroundMode.Checked;
			_worker.IsBackground = this.backgroundMode.Checked;
		}
		
		private DateTime _progressStart;
		private string _lastCheck;
		private void StartProgressTimer()
		{
			_progressStart = DateTime.Now;
			_lastCheck = null;
			_trendKnown = false;
			_originalTitle = null;
			UpdateProgressTimer();
			this.progressTimer.Enabled = true;
		}
		
		private void StopProgressTimer()
		{
			this.progressTimer.Enabled = false;
			UpdateProgressTimer();
			this.statusProgress.Visible = false;
		}
		
		private class Averager
		{
			private double _sum = 0;
			private double _count = 0;
			private double _value = 0;
			
			public void Add(double value)
			{
				_sum = _sum + value;
				_count = _count + 1;
				_value = _sum / _count;
			}
			
			public double Value
			{
				get
				{
					return _value;
				}
			}
		}
		
		private bool _trendKnown;
		private Averager _trendSeconds;
		private Averager _trendProgress;
		private string _originalTitle;
		
		private void UpdateProgressTimer()
		{
			int seconds = SecondsFrom(_progressStart);
			string elapsed = T2S(seconds);
			this.statusTimer.Text = elapsed;
			
			if (elapsed != _lastCheck)
			{
				_lastCheck = elapsed;
			
				string statusText = null;
				
				if (this.statusProgress.Visible && (this.statusProgress.Style != ProgressBarStyle.Marquee) && (this.statusProgress.Minimum != this.statusProgress.Maximum))
				{
					double progress = 100.0 * ((double)(this.statusProgress.Value - this.statusProgress.Minimum) / (double)(this.statusProgress.Maximum - this.statusProgress.Minimum));
					string progressText = progress.ToString("##0");
					if ((progressText != "100") && (progressText != "0"))
					{
						if (!_trendKnown)
						{
							_trendSeconds = new Averager();
							_trendProgress = new Averager();
							_trendKnown = true;
						}
						
						_trendSeconds.Add(seconds);
						_trendProgress.Add(progress);
				
						statusText = String.Format("{0}, {1}% left",
						                          T2S(Convert.ToInt32((100.0 - progress) * (_trendSeconds.Value / _trendProgress.Value))),
						                          (100 - Convert.ToInt32(progressText)));
					}
				}

				if (_originalTitle == null)
				{
					_originalTitle = this.Text;
				}
				if (String.IsNullOrEmpty(statusText))
				{
					statusText = _originalTitle;
				}
				else
				{
					statusText = String.Format("{0} - {1}", statusText, _originalTitle);
				}
				if (this.Text != statusText)
				{
					this.Text = statusText;
				}
			}
			
			if (!this.statusProgress.Visible && (seconds >= 2))
			{
				this.statusProgress.Style = ProgressBarStyle.Marquee;
				this.statusProgress.Visible = true;
			}
		}
		
		void ProgressTimerTick(object sender, EventArgs e)
		{
			UpdateProgressTimer();
		}
		
		void FileListMenuSelectAllClick(object sender, EventArgs e)
		{
			ListViewHelper.SelectAll(this.fileList, true);
		}
		
		void MainFormDragEnter(object sender, DragEventArgs e)
		{
			ValidateIndexMenu();
			if (!Locked && e.Data.GetDataPresent(DataFormats.FileDrop) && this.menuIndexAddFile.Enabled)
			{
				e.Effect = DragDropEffects.Copy;
			}
			else
			{
				e.Effect = DragDropEffects.None;
			}
		}
		
		private class AddFilesTask: System.Windows.Forms.IdleTask
		{
			private MainForm _parent;
			private string[] _fileNames;
			public AddFilesTask(MainForm parent, string[] fileNames)
			{
				_parent = parent;
				_fileNames = fileNames;
			}
			
			public override void Execute()
			{
				_parent.InternalAddFiles(_fileNames);
			}
		}
		
		void MainFormDragDrop(object sender, DragEventArgs e)
		{
			if(!Locked && e.Data.GetDataPresent(DataFormats.FileDrop))
			{
				new AddFilesTask(this, (string[])e.Data.GetData(DataFormats.FileDrop));
			}
		}

		#region Maintainer stuff
		private Indexer.Engine.Worker _maintainer = null;			// background worker that helps index keeping up-to-date
		private IDisposable _maintainerPump = null;
		
		private void PumpMaintainer()
		{
			if (!Locked)
			{
				bool doContinue = false;
				try
				{
					doContinue = _maintainer.Execute();
				}
				catch(Exception)
				{
				}
				if (!doContinue)
				{
					_maintainer.Dispose();
					_maintainer = null;
					return;
				}
			}
			_maintainerPump = IdleTask.New(this.PumpMaintainer);
		}
		
		private void ConnectMaintainer()
		{
			if (_index == null)
			{
				return;
			}
			_maintainer = _index.GetWorker();
			if (_maintainer == null)
			{
				return;
			}
			_maintainerPump = IdleTask.New(this.PumpMaintainer);
		}
		
		private void DisconnectMaintainer()
		{
			if (_maintainer == null)
			{
				return;
			}
			_maintainer.Dispose();
			_maintainer = null;
			if (_maintainerPump != null)
			{
				_maintainerPump.Dispose();
				_maintainerPump = null;
			}
		}
		#endregion
		
		void goButtonClick(object sender, EventArgs e)
		{
			ActivateSearch((Search)((ToolStripSplitButton)sender).DropDownItems[0].Tag);
		}
		
		void FileListMenuFindReferencesClick(object sender, EventArgs e)
		{
			if (this.fileList.SelectedIndices.Count == 1)
			{
				InternalSearch(
					Index.SetCombineMode.None,
					System.IO.Path.GetFileName(_search.Results[this.fileList.SelectedIndices[0]]),
					"",
					false,
					true);
			}
			else
			{
				// TODO: estimate progress correctly
				Wait(true);
				try
				{
					string[] fileNames = new string[this.fileList.SelectedIndices.Count];
					int fileNameIndex = 0;
					foreach(int index in this.fileList.SelectedIndices)
					{
						string fileName = System.IO.Path.GetFileName(_search.Results[index]);
						bool addIt = true;
						for(int i = 0; i < fileNameIndex; i++)
						{
							if (String.Equals(fileNames[i], fileName, StringComparison.InvariantCultureIgnoreCase))
							{
								addIt = false;
								break;
							}
						}
						if (addIt)
						{
							fileNames[fileNameIndex] = fileName;
							fileNameIndex++;
						}
					}
					System.Array.Resize(ref fileNames, fileNameIndex);
					
					ClearResults(false);
					
					foreach(string fileName in fileNames)
					{
						InternalSearch(
							Index.SetCombineMode.Or,
							fileName,
							"",
							false,
							true);
					}
				}
				finally
				{
					Wait(false);
				}
			}
		}

		private class UniquePathList: System.Collections.Generic.List<string>
		{
			public UniquePathList()
			{
			}
			
			public UniquePathList(int capacity)
			{
				base.Capacity = capacity;
			}
			
			public new void Add(string value)
			{
				foreach(string item in this)
				{
					if (String.Compare(item, value, true) == 0)
					{
						return;
					}
				}
				base.Add(value);
			}
		}
		
		void FileListMenuPropertiesClick(object sender, EventArgs e)
		{
			try
			{
				try
				{
					MessageForm message = new MessageForm(this);
					message.Text = "Document Properties";
	
					Wait(true);
					try
					{
						int count = this.fileList.SelectedIndices.Count;
						bool useDocumentBytes = false;
						bool useUnpackedBytes = false;
						long documentBytes = 0;
						long unpackedBytes = 0;
						bool useMetrics = false;
						bool useDate = false;
						DateTime minWriteTime = new DateTime();
						DateTime maxWriteTime = new DateTime();
						UniquePathList containers = new UniquePathList();
						UniquePathList containerPaths = new UniquePathList(count);
						UniquePathList folders = new UniquePathList(count);
						bool useContainers = false;
						bool useCount = true;	// TODO: ? count > 1;
						
						message.AddGroup(N2S("File{1}", count));
						
						for(int i = 0; i < count; i++)
						{
							int itemIndex = this.fileList.SelectedIndices[i];
			
							string fileName = _search.Results[itemIndex];
							
							Explorer.Model.Entry info = _index.IO.GetFileInfo(fileName);
							Index.DocumentKind kind = _search.Results.GetItemKind(itemIndex);
							
							message.AddValue(String.Format("\"{0}\"", fileName));
							
							folders.Add(System.IO.Path.GetDirectoryName(fileName));
		
							if (info != null)
							{
								string container;
								string containerPath;
								_index.IO.GetContainerInfo(fileName, out container, out containerPath);
								
								if (container != null)
								{
									useContainers = true;
									containers.Add(container);
									containerPaths.Add(containerPath);
								}
								
								DateTime lastWriteTime = info.LastWriteTime;
								
								if (useDate)
								{
									if (DateTime.Compare(lastWriteTime, minWriteTime) < 0)
									{
										minWriteTime = lastWriteTime;
									}
									if (DateTime.Compare(lastWriteTime, maxWriteTime) > 0)
									{
										maxWriteTime = lastWriteTime;
									}
								}
								else
								{
									useDate = true;
									minWriteTime = lastWriteTime;
									maxWriteTime = lastWriteTime;
								}
								
								useMetrics = true;
								
								if ((kind & Index.DocumentKind.Indexed) != 0)
								{
									useDocumentBytes = true;
									documentBytes += info.Size;
								}
								else
								{
									useUnpackedBytes = true;
									unpackedBytes += info.Size;
								}
							}
						}
		
						message.AddGroup(N2S("Folder{1}", folders.Count));
						foreach(string item in folders)
						{
							if (item.EndsWith("\\"))
							{
								message.AddValue(String.Format("\"{0}\"", item));
							}
							else
							{
								message.AddValue(String.Format("\"{0}\\\"", item));
							}
						}
						
						if (useContainers)
						{
							message.AddGroup(N2S("Container{1}", containers.Count));
							foreach(string item in containers)
							{
								message.AddValue(String.Format("\"{0}\"", item));
							}
							
							message.AddGroup(N2S("Path{1}", containerPaths.Count));
							foreach(string item in containerPaths)
							{
								message.AddValue(String.Format("\"{0}\"", item));
							}
						}
						
						if (useDate)
						{
							string min = T2S(minWriteTime);
							string max = T2S(maxWriteTime);
							
							if (min == max)
							{
								message.Add("Updated", min);
							}
							else
							{
								message.Add("Updated", min + " ... " + max);
							}
						}
						
						if (useMetrics)
						{
							message.AddGroup("Metrics");
					
							if (useCount)
							{
								message.AddValue(N2S("{0} document{1}", count));
							}
							if (useDocumentBytes)
							{
								message.AddValue(N2S("{0} document byte{1}", documentBytes));
							}
							if (useUnpackedBytes)
							{
								message.AddValue(N2S("{0} binary byte{1}", unpackedBytes));
							}
						}
					}
					finally
					{
						Wait(false);
					}
						
					message.Show();
				}
				catch(Exception error)
				{
					ShowException(error, this.DefaultErrorHandler);
				}
			}
			finally
			{
				WriteLastError();
			}
		}
		
		void FileListMenuCompareClick(object sender, EventArgs e)
		{
			try
			{
				try
				{
					CompareFilesCommand comparer = new CompareFilesCommand();
					GetComparer(out comparer.FileName, out comparer.Arguments);
	
					comparer.LeftArgument = _search.Results[this.fileList.SelectedIndices[0]];
					comparer.RightArgument = _search.Results[this.fileList.SelectedIndices[1]];
					comparer.IO = _index.IO;
					comparer.TemporaryFiles = _temporaryFiles;
					_worker.Execute(comparer);
				}
				catch(Exception error)
				{
					ShowException(error, this.DefaultErrorHandler);
				}
			}
			finally
			{
				WriteLastError();
			}
		}
		
		#region CustomDrawn list view
		// this code fixes "when controls are locked, current selection is not visible - whole list is painted in one color"
		// by using CustomDraw when list view is disabled.
		// this code was inspired by ObjectListView (https://sourceforge.net/project/platformdownload.php?group_id=225207)
		private class CustomDrawContext
		{
			public Graphics Graphics;
			public Rectangle Bounds;
			public ListViewItem Item;
			public ListViewItemStates ItemState;
		}
		
		void FileListDrawItem(object sender, DrawListViewItemEventArgs e)
		{
			e.DrawDefault = e.Item.ListView.Enabled;
		}

		private Color GetItemBackgroundColor(CustomDrawContext context)
		{
			ListView view = context.Item.ListView;
			
            if (!view.Enabled)
            {
                return SystemColors.Control;
            }
            return view.BackColor;
		}
		
		private Color GetItemTextBackgroundColor(CustomDrawContext context)
		{
			ListView view = context.Item.ListView;
			
            if (context.Item.Selected)
            {
            	return SystemColors.Highlight;
            }
            if (!view.Enabled)
            {
                return SystemColors.Control;
            }
            return view.BackColor;
		}
		
		private Color GetItemTextForegroundColor(CustomDrawContext context)
		{
			if (context.Item.Selected)
			{
				if (context.Item.ListView.Enabled)
				{
					return SystemColors.HighlightText;
				}
				else
				{
					return SystemColors.HighlightText;
				}
			}
			if (!context.Item.ListView.Enabled)
			{
				return SystemColors.GrayText;
			}
			return context.Item.ForeColor;
		}
		
		private void DrawListItemBackground(CustomDrawContext context)
		{
            Color backgroundColor = GetItemBackgroundColor(context);

            using (Brush brush = new SolidBrush(backgroundColor)) {
                context.Graphics.FillRectangle(brush, context.Bounds);
            }
		}
		
		private void DrawListItemTextBackground(CustomDrawContext context)
		{
            Color backgroundColor = GetItemTextBackgroundColor(context);

            using (Brush brush = new SolidBrush(backgroundColor)) {
                context.Graphics.FillRectangle(brush, context.Bounds);
            }
		}
		
        [DllImport("comctl32.dll", CharSet = CharSet.Auto)]
        private static extern bool ImageList_Draw(IntPtr himl, int i, IntPtr hdcDst, int x, int y, int fStyle);
        
        private const int ILD_NORMAL = 0x00000000;
        private const int ILD_TRANSPARENT = 0x00000001;
        private const int ILD_MASK = 0x00000010;
        private const int ILD_IMAGE = 0x00000020;
        private const int ILD_BLEND25 = 0x00000002;
        private const int ILD_BLEND50 = 0x00000004;

        public static bool DrawImageList(Graphics g, ImageList il, int index, int x, int y, bool isSelected)
        {
            int flags = ILD_TRANSPARENT;
            if (isSelected)
                flags |= ILD_BLEND50;
            bool result = ImageList_Draw(il.Handle, index, g.GetHdc(), x, y, flags);
            g.ReleaseHdc();
            return result;
        }
        
		private void DrawAlignedListItemImageAndText(CustomDrawContext context)
		{
			int imageIndex = context.Item.ImageIndex;
			if (imageIndex >= 0)
			{
				if (context.Item.Selected)
				{
					// paint the image "blended"
					DrawImageList(context.Graphics, context.Item.ListView.SmallImageList, imageIndex, context.Bounds.Left, context.Bounds.Top, true);
				}
				else
				{
					context.Item.ListView.SmallImageList.Draw(context.Graphics, new Point(context.Bounds.Left, context.Bounds.Top), imageIndex);
				}
			}
			int offset = context.Item.ListView.SmallImageList.ImageSize.Width;
			context.Bounds.X += offset - 1 /* TODO: ??? */;
			context.Bounds.Width -= offset - 2 /* TODO: ??? */;
			
			DrawListItemTextBackground(context);
			
			string text = context.Item.SubItems[0].Text;
			
            TextFormatFlags flags = TextFormatFlags.EndEllipsis | TextFormatFlags.NoPrefix |
                TextFormatFlags.VerticalCenter | TextFormatFlags.PreserveGraphicsTranslateTransform;
            TextRenderer.DrawText(context.Graphics, text, context.Item.ListView.Font, context.Bounds, GetItemTextForegroundColor(context), Color.Transparent, flags);
		}
		
		void FileListDrawSubItem(object sender, DrawListViewSubItemEventArgs e)
		{
			e.DrawDefault = false;
			
			CustomDrawContext context = new CustomDrawContext();
			context.Graphics = e.Graphics;
			context.Bounds = e.Bounds;
			context.Item = e.Item;
			context.ItemState = e.ItemState;
			
			DrawListItemBackground(context);

            // Adjust the first columns rectangle to match the padding used by the native mode of the ListView
            {
                context.Bounds.X += 4;
                context.Bounds.Width -= 4;
            }
            DrawAlignedListItemImageAndText(context);
		}
		#endregion
	}
}
