using System;
using Indexer.Engine;

namespace Indexer.Gui
{
	public class Search
	{
		private Index.DocumentCollection _results = null;
		private Index.Query _query = null;
		private string _masks = null;
		private System.Collections.BitArray _selected = null;
		
		public bool CaseSensitive;
		public bool WholeWords;
		public string Phrase;

		public void Dispose()
		{
			// TODO: call Dispose() in members
			_results = null;
			_query = null;
			_masks = null;
			_selected = null;
		}
		
		public void SetSelectionFromListView(System.Windows.Forms.ListView view)
		{
			// don't override selection if already recorded
			if (_selected == null)
			{
				_selected = new System.Collections.BitArray(_results.Count);
				foreach(int index in view.SelectedIndices)
				{
					_selected[index] = true;
				}
			}
		}
		
		public void SetSelectionToListView(System.Windows.Forms.ListView view)
		{
			int focusedIndex = -1;
			view.SelectedIndices.Clear();
			if (_selected != null)
			{
				for(int index = 0; index < _selected.Count; index++)
				{
					if (_selected[index])
					{
						if (focusedIndex < 0)
						{
							focusedIndex = index;
						}
						view.SelectedIndices.Add(index);
					}
				}
			}
			if (focusedIndex >= 0)
			{
				System.Windows.Forms.ListViewItem item = view.Items[focusedIndex];
				item.Focused = true;
				item.Selected = true;
				// TODO: somehow hide the default focus rectangle
				item.EnsureVisible();
			}
		}
		
		public Search(Index.Query query)
		{
			Initialize(query, null, null);
		}
		
		public Search(Index.Query query, string masks)
		{
			Initialize(query, masks, null);
		}

		public Search(Index.Query query, string masks, Index.DocumentCollection results)
		{
			Initialize(query, masks, results);
		}
		
		private void Initialize(Index.Query query, string masks, Index.DocumentCollection results)
		{
			_query = query;
			_masks = masks;
			_results = results;
		}
		
		public Index.DocumentCollection Results
		{
			get
			{
				return _results;
			}
		}
		
		public Index.Query Query
		{
			get
			{
				return _query;
			}
		}
		
		public string Masks
		{
			get
			{
				return _masks;
			}
		}

		private string GetDisplayLabelTailForResults()
		{
			if (Results.Count == 0)
			{
				return "no results";
			}
			else if (Results.Count == 1)
			{
				return "1 result";	
			}
			else
			{
				
				return String.Format("{0} results", Results.Count.ToString("#,###,###,##0"));
			}
		}
		
		private string GetDisplayLabelHeadForPhrase(string phrase)
		{
			if (CaseSensitive)
			{
				if (WholeWords)
				{
					return String.Format(
						"Case sensitive search for phrase {0}",
						phrase);
				}
				else
				{
					return String.Format(
						"Case sensitive search for {0}",
						phrase);
				}
			}
			else
			{
				if (WholeWords)
				{
					return String.Format(
						"Search for phrase {0}",
						phrase);
				}
				else
				{
					return String.Format(
						"Search for {0}",
						phrase);
				}
			}
		}
		
		private string GetDisplayLabelForMasksAndPhrase(string masks, string phrase)
		{
			return String.Format(
				"{0} in {1} returned {2}",
				GetDisplayLabelHeadForPhrase(phrase), masks, GetDisplayLabelTailForResults());
		}
		
		private string GetDisplayLabelForMasks(string masks)
		{
			return String.Format(
				"Find {0} returned {1}",
				masks, GetDisplayLabelTailForResults());
		}
		
		private string GetDisplayLabelForPhrase(string phrase)
		{
			return String.Format(
				"{0} returned {1}",
				GetDisplayLabelHeadForPhrase(phrase), GetDisplayLabelTailForResults());
		}
		
		private string GetDisplayLabel()
		{
			return String.Format(
				"List returned {0}",
				GetDisplayLabelTailForResults());
		}
		
		public string DisplayLabel
		{
			get
			{
				string phrase = "\"" + String.Join("\",\"", Query.Phrases) + "\"";
				string masks = Masks;				// TODO: use masks from query
				
				bool phraseUsed = phrase != "\"\"";
				bool masksUsed = !String.IsNullOrEmpty(masks) && (masks != "*") && (masks != "*.*");
				
				if (masksUsed)
				{
					masks = "\"" + masks + "\"";
					if (phraseUsed)
					{
						return GetDisplayLabelForMasksAndPhrase(masks, phrase);
					}
					else
					{
						return GetDisplayLabelForMasks(masks);
					}
				}
				else
				{
					if (phraseUsed)
					{
						return GetDisplayLabelForPhrase(phrase);
					}
					else
					{
						return GetDisplayLabel();
					}
				}
			}
		}
	}

	public class SearchHistory
	{
		private System.Collections.Generic.List<Search> _items = new System.Collections.Generic.List<Search>();
		
		public int IndexOf(Search item)
		{
			return _items.IndexOf(item);
		}
		
		public void InsertAfter(Search item, Search anchor)
		{
			int index;
			if (anchor == null)
			{
				index = Count - 1;
			}
			else
			{
				index = IndexOf(anchor);
				if (index < 0)
				{
					index = Count - 1;
				}
			}
			_items.Insert(index + 1, item);
		}
		
		public int Count
		{
			get
			{
				return _items.Count;
			}
		}
		
		public Search this[int index]
		{
			get
			{
				return _items[index];
			}
		}
		
		public void Remove(Search item)
		{
			int index = IndexOf(item);
			if (index >= 0)
			{
				_items.RemoveAt(index);
				item.Dispose();
			}
		}
		
		public void Clear()
		{
			while (Count > 0)
			{
				Remove(this[0]);
			}
		}
	}
}
