using System;

namespace Explorer.Engine
{
	using TreeId = System.Int32;
	using LeafId = System.Int32;
	
	public class NameTree
	{
		private class Names: System.Collections.Generic.SortedDictionary<string, NameTree>
		{
			
		}
		
		private Names _items = null;
		private Names Items
		{
			get
			{
				if (_items == null)
				{
					_items = new Names();
				}
				return _items;
			}
		}
		
		protected void Initialize()
		{
			if (_items == null)
			{
				_items = new Names();
			}
		}
		
		public NameTree()
		{
		}
		
		private string _name = null;
		public string Name
		{
			get
			{
				return _name;
			}
			protected set
			{
				_name = value;
			}
		}
		
		public char PathSeparator = '/';

		private void Split(string fullName, out string folderName, out string entryName)
		{
			if (String.IsNullOrEmpty(fullName))
			{
				folderName = "";
				entryName = "";
			}
			else
			{
				int i = fullName.LastIndexOf(PathSeparator);
				if (i < 0)
				{
					folderName = "";
					entryName = fullName;
				}
				else
				{
					folderName = fullName.Substring(0, i);
					entryName = fullName.Substring(i + 1);
				}
			}
		}
		
		private static string Normalized(string name)
		{
			return name.ToUpper();
		}
		
		protected NameTree SimpleFind(string name)
		{
			if (String.IsNullOrEmpty(name))
			{
				return this;
			}
			
			if (_items == null)
			{
				return null;
			}

			NameTree result;
			if (_items.TryGetValue(Normalized(name), out result))
			{
				return result;
			}
			else
			{
				return null;
			}
		}
		
		public NameTree Find(string name)
		{
			if (String.IsNullOrEmpty(name))
			{
				return this;
			}
			if (_items == null)
			{
				return null;		
			}
			else
			{
				string folderName;
				string entryName;
				Split(name, out folderName, out entryName);
				
				NameTree home;
				if (String.IsNullOrEmpty(folderName))
				{
					home = this;
				}
				else
				{
					home = Find(folderName);
				}
				
				if (home == null)
				{
					return null;
				}
				else
				{
					return home.SimpleFind(entryName);
				}
			}
		}

		private class NameEnumerator: System.Collections.Generic.IEnumerable<string>, System.Collections.Generic.IEnumerator<string>
		{
			private System.Collections.Generic.IEnumerator<System.Collections.Generic.KeyValuePair<string, NameTree>> _enumerator;
			
			public NameEnumerator(System.Collections.Generic.IEnumerator<System.Collections.Generic.KeyValuePair<string, NameTree>> enumerator)
			{
				_enumerator = enumerator;
				_enumerator.Reset();
			}
			
			public System.Collections.Generic.IEnumerator<string> GetEnumerator()
			{
				return this;
			}
			
			System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
			{
				return this;
			}
			
			public string Current 
			{
				get 
				{
					return _enumerator.Current.Value.Name;	
				}
			}
			
			object System.Collections.IEnumerator.Current 
			{
				get 
				{
					return _enumerator.Current.Value.Name;
				}
			}
			
			public void Dispose()
			{
				if (_enumerator != null)
				{
					_enumerator.Dispose();
					_enumerator = null;
				}
			}
			
			public bool MoveNext()
			{
				return _enumerator.MoveNext();
			}
			
			public void Reset()
			{
				_enumerator.Reset();
			}
		}
		
		public System.Collections.Generic.IEnumerable<string> GetNames()
		{
			return new NameEnumerator(Items.GetEnumerator());
		}
			
		public TreeId Id;
		
		private NameTree Parent;
		
		public string FullName
		{
			get
			{
				if (Parent == null)
				{
					return Name;
				}
				else
				{
					return Parent.FullName + PathSeparator + Name;
				}
			}
		}
		
		protected virtual TreeId CreateId()
		{
			if (Parent == null)
			{
				return 0;
			}
			else
			{
				return Parent.CreateId();
			}
		}
		
		public NameTree Add(string name)
		{
			string folderName;
			string entryName;
			Split(name, out folderName, out entryName);
			
			NameTree home;
			if (String.IsNullOrEmpty(folderName))
			{
				home = this;
			}
			else
			{
				home = Add(folderName);
			}
			
			NameTree item = home.SimpleFind(entryName);
			if (item == null)
			{
				item = new NameTree();
				item.PathSeparator = home.PathSeparator;
				item.Name = entryName;
				item.Parent = home;
				item.Id = home.CreateId();
				home.Items.Add(Normalized(entryName), item);
				return item;
			}
			return item;
		}
	}

	public class NameTreeRoot: NameTree
	{
		public NameTreeRoot()
		{
			// this MUST be called so that the _items member is not empty
			// especially needed for call to Find("/") - which is valid for
			// the root tree but would fail without _items being initialized
			base.Initialize();
		}
		
		private TreeId _idSeed;
		
		protected override TreeId CreateId()
		{
			return ++_idSeed;
		}
	}

	public class NameTreeLeafMap
	{
		private static LeafId OuterToInner(LeafId id)
		{
			return id + 1;
		}
		
		private static LeafId InnerToOuter(LeafId id)
		{
			return id - 1;
		}
		
		protected struct LeafEntry
		{
			public TreeId Tree;
			public LeafId Leaf;
		}

		private class Comparer: System.Collections.Generic.Comparer<LeafEntry>
		{
			private NameTreeLeafMap _parent;
			private string _nullName;
			
			public Comparer(NameTreeLeafMap parent, string nullName)
			{
				_parent = parent;
				_nullName = nullName;
			}
			
			public Comparer(NameTreeLeafMap parent)
			{
				_parent = parent;
			}

			private string GetName(LeafId id)
			{
				if (id == 0)
				{
					return _nullName;
				}
				else
				{
					return _parent.GetName(InnerToOuter(id));
				}
			}
			
			public override int Compare(LeafEntry x, LeafEntry y)
			{
				int result = x.Tree - y.Tree;
				if (result == 0)
				{
					return String.Compare(GetName(x.Leaf), GetName(y.Leaf));
				}
				else
				{
					return result;
				}
			}
		}
		
		private int BinarySearch(TreeId treeId, string leafName)
		{
			Comparer comparer = new Comparer(this, leafName);
			if (!_sorted)
			{
				_sorted = true;
				_map.Sort(comparer);
			}
			LeafEntry entry = new LeafEntry();
			entry.Tree = treeId;
			entry.Leaf = 0;
			return _map.BinarySearch(entry, comparer);
		}
		
		private int BinarySearch(TreeId treeId)
		{
			return BinarySearch(treeId, String.Empty);
		}
		
		public void Prepare()
		{
			// TODO: this prepare-phase can be delayed until needed later but it may cause unexpected slow-down when the sorting is hit for the first time
			if (!_sorted)
			{
				_sorted = true;
				_map.Sort(new Comparer(this));
			}
		}
		
		private System.Collections.Generic.List<LeafEntry> _map = null;
		private bool _sorted = false;
		public NameTreeLeafMap(int capacity)
		{
			_map = new System.Collections.Generic.List<LeafEntry>(capacity);
		}
		
		public int Count
		{
			get
			{
				return _map.Count;
			}
		}
		
		private void Add(TreeId treeId, LeafId leafId)
		{
			LeafEntry item = new LeafEntry();
			item.Tree = treeId;
			item.Leaf = OuterToInner(leafId);
			_map.Add(item);
			_sorted = false;
		}

		#region Main interface
		public delegate string NameProvider(LeafId id);
		public NameProvider GetName;
		
		public System.Collections.Generic.IEnumerable<LeafId> GetEnumerator(NameTree tree)
		{
			return GetEnumerator(tree.Id);
		}
		
		public void Add(NameTree tree, LeafId leafId)
		{
			Add(tree.Id, leafId);
		}
		
		public bool Find(NameTree tree, string leafName, out LeafId leafId)
		{
			if (tree == null)
			{
				leafId = 0;
				return false;
			}
			
			int index = BinarySearch(tree.Id, leafName);
			if (index < 0)
			{
				leafId = 0;
				return false;
			}
			else
			{
				leafId = InnerToOuter(_map[index].Leaf);
				return true;
			}
		}
		#endregion
		
		private class LeafEnumerator: System.Collections.Generic.IEnumerable<LeafId>, System.Collections.Generic.IEnumerator<LeafId>
		{
			private NameTreeLeafMap _parent;
			private int _firstIndex;
			private int _index;
			private bool _initialized;
			private TreeId _treeId;
			
			public LeafEnumerator(NameTreeLeafMap parent, TreeId treeId)
			{
				_parent = parent;
				_treeId = treeId;
				
				int index = _parent.BinarySearch(_treeId);
				if (index < 0)
				{
					index = ~index;
				}
				_firstIndex = index;
				Reset();
			}
			
			public System.Collections.Generic.IEnumerator<int> GetEnumerator()
			{
				return this;
			}
			
			System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
			{
				return this;
			}
			
			public int Current 
			{
				get 
				{
					return NameTreeLeafMap.InnerToOuter(_parent._map[_index].Leaf);
				}
			}
			
			object System.Collections.IEnumerator.Current 
			{
				get 
				{
					return NameTreeLeafMap.InnerToOuter(_parent._map[_index].Leaf);
				}
			}
			
			public void Dispose()
			{
			}
			
			public bool MoveNext()
			{
				if (!_initialized)
				{
					_initialized = true;
					_index = _firstIndex;
				}
				else
				{
					_index++;
				}
				
				if (_index < _parent._map.Count)
				{
					if (_parent._map[_index].Tree == _treeId)
					{
						return true;
					}
				}
				return false;
			}
			
			public void Reset()
			{
				_initialized = false;
			}
		}
		
		private System.Collections.Generic.IEnumerable<LeafId> GetEnumerator(TreeId treeId)
		{
			return new LeafEnumerator(this, treeId);
		}
	}
}
