using System;

namespace Explorer.Engine
{
	public abstract class TreeMappedFileSystem: FileSystemBase
	{
		private string _drive;
		protected const string _innerRoot = "/";
		
		private NameTree _folders = null;
		private NameTreeLeafMap _files = null;		
		
		public TreeMappedFileSystem(char drive)
		{
			_drive = drive.ToString().ToUpper() + ":\\";
		}
		
		protected char Drive
		{
			get
			{
				return _drive[0];
			}
		}

		protected void InitializeFolders()
		{
			_folders = new NameTreeRoot();
			_folders.PathSeparator = '/';
		}
		
		protected void InitializeFiles(int capacity, NameTreeLeafMap.NameProvider nameProvider)
		{
			_files = new NameTreeLeafMap(capacity);
			_files.GetName = nameProvider;
		}
		
		private bool _opened = false;
		public sealed override void Open()
		{
			if (_opened)
			{
				return;
			}
			_opened = true;
			InitializeFolders();
			InitializeFiles(0, null);
			BuildIndexes();
			Files.Prepare();
		}
		
		protected void RegisterFolder(string folder)
		{
			folder = OuterToInner(folder);
			if (folder.EndsWith("/"))
			{
				folder = folder.Substring(0, folder.Length - 1);
			}
			if (!String.IsNullOrEmpty(folder))
			{
				Folders.Add(folder);
			}
		}
		
		private static string CleanZipName(string name)
		{
			if (name.StartsWith("/"))
			{
				return name.Substring(1);
			}
			else
			{
				return name;
			}
		}
	
		#region To be implemented in derived classes
		protected virtual void AddFilesToList(NameTree rootTree, Model.Entries result)
		{
		}
		
		protected virtual Explorer.Model.Entry GetFileInfo(string innerName, int fileIndex)
		{
			return null;
		}
		
		protected virtual void BuildIndexes()
		{
		}
		
		protected virtual System.IO.Stream OpenFile(Model.Entry info, System.IO.FileMode mode, System.IO.FileAccess access, System.IO.FileShare share)
		{
			throw new NotImplementedException();
		}
		#endregion
		
		public long FileCount
		{
			get
			{
				return Files.Count;
			}
		}
		
		protected NameTree Folders
		{
			get
			{
				Open();
				return _folders;
			}
		}
		
		protected NameTreeLeafMap Files
		{
			get
			{
				Open();
				return _files;
			}
		}
		
		protected override void Close()
		{
			base.Close();
			_folders = null;
			_files = null;
		}
		
		private Model.Entries CreateList()
		{
			Model.Entries result = new Model.Entries();
			Model.Entry item = new Model.Entry();
			item.Kind = Model.EntryKind.Folder;
			item.Name = _drive;
			result.Add(item);
			return result;
		}
		
		protected string OuterToInner(string path)
		{
			return _innerRoot + path.Substring(_drive.Length).Replace("\\", "/");
		}
		
		protected string InnerToOuter(string path)
		{
			return _drive + path.Substring(_innerRoot.Length).Replace("/", "\\");
		}
		
		public sealed override Explorer.Model.Entries CreateList(string folder)
		{
			if (String.IsNullOrEmpty(folder))
			{
				return CreateList();
			}
			
			folder = OuterToInner(folder);
			if (!folder.EndsWith("/"))
			{
				folder = folder + "/";
			}
			
			Model.Entries result = new Model.Entries();
			
			{
				Model.Entry item = new Model.Entry();
				item.Kind = Model.EntryKind.UpFolder;
				item.Path = InnerToOuter(folder);
				result.Add(item);
			}

			NameTree rootTree = Folders.Find(folder.Substring(0, folder.Length - 1));
			if (rootTree == null)
			{
				return result;
			}
			
			foreach(string folderName in rootTree.GetNames())
			{
				result.Add(GetFolderInfo(folder + folderName, false));
			}
			
			AddFilesToList(rootTree, result);
			
			return result;
		}
		
		protected static string GetPath(string fullName)
		{
			if (String.IsNullOrEmpty(fullName))
			{
				return "";
			}
			
			if (fullName.EndsWith(":\\"))
			{
				return "";
			}
			
			if (fullName.EndsWith("\\"))
			{
				fullName = fullName.Substring(0, fullName.Length - 1);
			}
			
			int i = fullName.LastIndexOf("\\");
			if (i < 0)
			{
				return "";
			}
			else
			{
				return fullName.Substring(0, i + 1);
			}
		}

		private Model.Entry GetFolderInfo(string innerFolderName, bool checkIfExists)
		{
			if (checkIfExists)
			{
				if (Folders.Find(innerFolderName) == null)
				{
					return null;
				}
			}
			
			Model.Entry item = new Model.Entry();
			item.Kind = Model.EntryKind.Folder;
			item.FullName = InnerToOuter(innerFolderName);
			
			return item;
		}
		
		public sealed override Explorer.Model.Entry GetInfo(string entryName)
		{
			entryName = OuterToInner(entryName);
			if (String.IsNullOrEmpty(entryName))
			{
				return null;
			}
			if (entryName.EndsWith("/"))
			{
				return GetFolderInfo(entryName, true);
			}
			else
			{
				Model.Entry result = GetFolderInfo(entryName, true);
				if (result == null)
				{
					int j = entryName.LastIndexOf("/");
					if (j < 0)
					{
						// TODO: should not ever happen
						return null;
					}
					int index;
					if (Files.Find(Folders.Find(entryName.Substring(0, j)), entryName.Substring(j + 1), out index))
					{
						return GetFileInfo(entryName, index);
					}
					else
					{
						return null;
					}
				}
				else
				{
					return result;
				}
			}
		}
		
		public sealed override System.IO.Stream OpenFile(string fileName, System.IO.FileMode mode, System.IO.FileAccess access, System.IO.FileShare share)
		{
			Model.Entry info = GetInfo(fileName);
			if ((info == null) || (info.InternalHandle == null) || (info.Kind != Model.EntryKind.File))
			{
				throw new System.IO.FileNotFoundException();
			}
			return OpenFile(info, mode, access, share);
		}
	}
}
