using System;

namespace Explorer.Model
{
	public class Reader
	{
		private System.LocalDataStoreSlot _contextSlot = System.Threading.Thread.AllocateDataSlot();
		private Engine.IFileSystem _fileSystem = null;
		
		public Reader()
		{
		}
		
		public Engine.IFileSystem FileSystem
		{
			get
			{
				return _fileSystem;
			}
			set
			{
				DisconnectThread();
				if (_fileSystem != value)
				{
					if (_fileSystem != null)
					{
						_fileSystem.Release();
					}
					_fileSystem = value;
					if (_fileSystem != null)
					{
						_fileSystem.AddRef();
						_fileSystem.Open();
					}
				}
			}
		}
		
		private class Context
		{
			internal string _getListPath = null;
			internal Entries _getListResults = null;
			internal string _getInfoPath = null;
			internal Entry _getInfoResult = null;
			
			public void Close()
			{
				if (_getListResults != null)
				{
					_getListResults.Dispose();
					_getListResults = null;
				}
				_getListPath = null;
				if (_getInfoResult != null)
				{
					_getInfoResult.Dispose();
					_getInfoResult = null;
				}
				_getInfoPath = null;
			}
		}
		
		private Context GetThreadContext()
		{
			Context result = (Context)System.Threading.Thread.GetData(_contextSlot);
			if (result == null)
			{
				result = new Context();
				System.Threading.Thread.SetData(_contextSlot, result);
			}
			return result;
		}
		
		public void DisconnectThread()
		{
			lock(this)
			{
				Context ctx = (Context)System.Threading.Thread.GetData(_contextSlot);
				if (ctx != null)
				{
					ctx.Close();
					ctx = null;
					System.Threading.Thread.SetData(_contextSlot, null);
				}
			}
		}
		
		private Entry InternalGetInfo(Context ctx, string path, Model.EntryKind kind)
		{
			if (_fileSystem == null)
			{
				return null;
			}
			Entry info = null;
			
			if (ctx._getInfoPath == path)
			{
				info = ctx._getInfoResult;
			}
			else
			{
				if (ctx._getInfoResult != null)
				{
					ctx._getInfoResult.Dispose();
					ctx._getInfoResult = null;
				}
				ctx._getInfoPath = path;
				ctx._getInfoResult = _fileSystem.GetInfo(path);
				info = ctx._getInfoResult;
			}
			
			if (info == null)
			{
				return null;
			}
			if (info.Kind != kind)
			{
				return null;
			}
			return info;
		}
		
		public Entry GetFileInfo(string path)
		{
			lock(this)
			{
				return InternalGetInfo(GetThreadContext(), path, EntryKind.File);
			}
		}
		
		public Entry GetDirectoryInfo(string path)
		{
			lock(this)
			{
				return InternalGetInfo(GetThreadContext(), path, EntryKind.Folder);
			}
		}
		
		public bool FileExists(string path)
		{
			Entry info = GetFileInfo(path);
			if (info == null)
			{
				return false;
			}
			else
			{
				return true;
			}
		}
		
		public bool DirectoryExists(string path)
		{
			Entry info = GetDirectoryInfo(path);
			if (info == null)
			{
				return false;
			}
			else
			{
				return true;
			}
		}
		
		public System.IO.Stream OpenFile(string path, System.IO.FileMode mode, System.IO.FileAccess access, System.IO.FileShare share)
		{
			if (_fileSystem == null)
			{
				throw new System.IO.FileNotFoundException();
			}
			lock(this)
			{
				return _fileSystem.OpenFile(path, mode, access, share);
			}
		}
		
		public System.IO.TextReader CreateTextReader(string path)
		{
			if (_fileSystem == null)
			{
				throw new System.IO.FileNotFoundException();
			}
			lock(this)
			{
				System.IO.Stream stream = OpenFile(path, System.IO.FileMode.Open, System.IO.FileAccess.Read, System.IO.FileShare.Read);
				try
				{
					System.Text.Encoding encoding = System.Text.FileStream.GetEncoding(stream);
					return new System.IO.StreamReader(stream, encoding);
				}
				catch(Exception)
				{
					stream.Dispose();
					throw;
				}
			}
		}
		
		private System.Collections.Generic.ICollection<string> InternalGetEntries(Context ctx, string path, Model.EntryKind kind)
		{
			if (_fileSystem == null)
			{
				return new System.Collections.Generic.List<string>();
			}
			
			Entries list = null;
			
			if (ctx._getListPath == path)
			{
				list = ctx._getListResults;
			}
			else
			{
				if (ctx._getListResults != null)
				{
					ctx._getListResults.Dispose();
					ctx._getListResults = null;
				}
				ctx._getListPath = path;
				ctx._getListResults = _fileSystem.CreateList(path);
				list = ctx._getListResults;
			}
			
			System.Collections.Generic.List<string> result = new System.Collections.Generic.List<string>(list.Count);
			foreach(Entry info in list)
			{
				if (info.Kind == kind)
				{
					result.Add(info.FullName);
				}
			}
			return result;
		}
		
		public System.Collections.Generic.ICollection<string> GetFiles(string path)
		{
			lock(this)
			{
				return InternalGetEntries(GetThreadContext(), path, Model.EntryKind.File);
			}
		}
		
		public System.Collections.Generic.ICollection<string> GetDirectories(string path)
		{
			lock(this)
			{
				return InternalGetEntries(GetThreadContext(), path, Model.EntryKind.Folder);
			}
		}
		
		public bool IsFileNative(string path)
		{
			if (_fileSystem == null)
			{
				return false;
			}
			Explorer.Engine.IRawFileSystem baseFileSystem;
			string baseFileName = _fileSystem.GetBaseFileName(path, out baseFileSystem);
			if (String.IsNullOrEmpty(baseFileName))
			{
				// file not found
				return false;
			}
			return baseFileSystem == null;
		}
		
		public void GetContainerInfo(string path, out string container, out string containerPath)
		{
			if (_fileSystem == null)
			{
				container = null;
				containerPath = null;
				return;
			}
			Explorer.Engine.IRawFileSystem baseFileSystem;
			string baseFileName = _fileSystem.GetBaseFileName(path, out baseFileSystem);
			if (baseFileSystem == null)
			{
				// native file
				container = null;
				containerPath = baseFileName;
			}
			else
			{
				Explorer.Engine.IFileSystem fs = baseFileSystem as Explorer.Engine.IFileSystem;
				if (fs != null)
				{
					fs.GetContainerInfo(baseFileName, out container, out containerPath);
				}
				else
				{
					// TODO: implement
					throw new NotSupportedException();
				}
			}
		}
		
		public void CopyFile(string path, string targetNativePath)
		{
			CopyFile(path, targetNativePath, null);
		}
		
		public void CopyFile(string path, string targetNativePath, System.Threading.IProgress observer)
		{
			bool useObserver = observer != null;
			
			byte[] buffer = new byte[4 /* TODO: TUNE: */ * 1024];
			System.IO.Stream reader = OpenFile(path, System.IO.FileMode.Open, System.IO.FileAccess.Read, System.IO.FileShare.Read);
			try
			{
				System.IO.Directory.CreateDirectory(System.IO.Path.GetDirectoryName(targetNativePath));
				System.IO.Stream writer = new System.IO.FileStream(targetNativePath, System.IO.FileMode.Create, System.IO.FileAccess.Write);
				try
				{
					if (useObserver)
					{
						observer.Open(0, (ulong)reader.Length);
					}
					ulong count = 0;
					try
					{
						while (true)
						{
							int size = reader.Read(buffer, 0, buffer.Length);
							if (size > 0)
							{
								writer.Write(buffer, 0, size);
								if (useObserver)
								{
									count = count + (ulong)size;
									observer.Step(count);
								}
							}
							else
							{
								break;
							}
						}
						
						writer.Close();
						System.IO.File.SetLastWriteTime(targetNativePath, GetFileInfo(path).LastWriteTime);
					}
					finally
					{
						if (useObserver)
						{
							observer.Close();
						}
					}
				}
				finally
				{
					writer.Dispose();
				}
			}
			finally
			{
				reader.Dispose();
			}
		}
	}
}
