using System;
using System.Collections;
using System.Drawing;

namespace Quest.Graph2D
{
	public class Cluster: IGraphElement
	{
		protected Rectangle pBounds;
		protected IList pNodes;
		protected IList pEdges;

		public Cluster(Rectangle aBounds)
		{
			pBounds = aBounds;
			pNodes = new ArrayList();
			pEdges = new ArrayList();
		}

		public IEnumerable Nodes
		{
			get
			{
				return pNodes;
			}
		}

		public IEnumerable Edges
		{
			get
			{
				return pEdges;
			}
		}

		public Rectangle Bounds
		{
			get
			{
				return pBounds;
			}
		}

		public void Insert(Node node)
		{
			pNodes.Add(node);
		}

		public void Insert(Edge edge)
		{
			pEdges.Add(edge);
		}

		public object HitTest(int aX, int anY)
		{
			if(!Bounds.Contains(aX, anY)) return null;
			Node candidate = null;
			foreach(Node node in pNodes)
			{
				if(node.HitTest(aX, anY) != null)
				{
					if(candidate == null)
					{
						candidate = node;
					}
					else if(candidate.Index < node.Index)
					{
						// Index reflects visual order. lower index means
						// lower level
						candidate = node;
					}
				}
			}
			if(candidate != null) return candidate;
			foreach(Edge edge in pEdges)
			{
				if(edge.HitTest(aX, anY) != null) return edge;
			}
			return null;
		}
	}

	public class XYArray
	{
		public int LengthX = 0;
		public int LengthY = 0;

		private IGraphElement[] pItems = null;

		public XYArray(int aLengthX, int aLengthY)
		{
			LengthX = aLengthX;
			LengthY = aLengthY;
			int size = aLengthX * aLengthY;
			if(size > 0)
			{
				pItems = new IGraphElement[size];
				Array.Clear(pItems, 0, size);
			}
		}

		public IGraphElement this[int aX, int anY]
		{
			get
			{
				return pItems[aX + anY * LengthX];
			}
			set
			{
				pItems[aX + anY * LengthX] = value;
			}
		}

		public IGraphElement this[Point index]
		{
			get
			{
				return this[index.X, index.Y];
			}
		}

		public XYArray Extend(int aLeftIncrement, int aTopIncrement, int aRightIncrement, int aBottomIncrement)
		{
			XYArray result = new XYArray
				(
				 aLeftIncrement + LengthX + aRightIncrement,
				 aTopIncrement + LengthY + aBottomIncrement
				);

			for(int x = 0; x < LengthX; x++)
				for(int y = 0; y < LengthY; y++)
			{
				result[aLeftIncrement + x, aTopIncrement + y] = this[x, y];
			}
			return result;
		}
	}

	public class Quadrant: IGraphElement
	{
		// TUNE:
		protected const int SubspaceWidth = 300;
		protected const int SubspaceHeight = 200;

		protected Rectangle pBounds;

		protected XYArray pItems = null;

		public Quadrant()
		{
		}

		public bool IsEmpty
		{
			get
			{
				return (pItems == null);
			}
		}

		public Rectangle Bounds
		{
			get
			{
				return pBounds;
			}
		}

		public virtual object HitTest(int aX, int anY)
		{
			if(IsEmpty) return null;
			if(!Bounds.Contains(aX, anY)) return null;

			IGraphElement sphere = pItems[SubspaceIndex(aX, anY)];
			return (sphere == null) ? null : sphere.HitTest(aX, anY);
		}

		protected Point SubspaceIndex(int aX, int anY)
		{
			return new Point
				(
				 (aX - Bounds.Left) / SubspaceWidth,
				 (anY - Bounds.Top) / SubspaceHeight
				);
		}

		protected Point SubspaceIndex(Point p)
		{
			return SubspaceIndex(p.X, p.Y);
		}

		protected Point NormalizedSubspaceIndex(int aX, int anY)
		{
			Point result = SubspaceIndex(aX, anY);
			result.X = Math.Max(0, Math.Min(pItems.LengthX - 1, result.X));
			result.Y = Math.Max(0, Math.Min(pItems.LengthY - 1, result.Y));
			return result;
		}

		protected void Cover(int aX, int anY)
		{
			if(IsEmpty)
			{
				pItems = new XYArray(1, 1);
				pBounds = new Rectangle(aX, anY, SubspaceWidth, SubspaceHeight);
			}
			else
			{
				Point index = SubspaceIndex(aX, anY);
				int leftIncrement;
				int rightIncrement;
				int topIncrement;
				int bottomIncrement;

				if(index.X < 0)
				{
					leftIncrement = -index.X;
					rightIncrement = 0;
				}
				else if(index.X >= pItems.LengthX)
				{
					leftIncrement = 0;
					rightIncrement = index.X - (pItems.LengthX - 1);
				}
				else
				{
					leftIncrement = 0;
					rightIncrement = 0;
				}

				if(index.Y < 0)
				{
					topIncrement = -index.Y;
					bottomIncrement = 0;
				}
				else if(index.Y >= pItems.LengthY)
				{
					topIncrement = 0;
					bottomIncrement = index.Y - (pItems.LengthY - 1);
				}
				else
				{
					topIncrement = 0;
					bottomIncrement = 0;
				}

				if(leftIncrement + rightIncrement + topIncrement + bottomIncrement > 0)
				{
					pItems = pItems.Extend(leftIncrement, topIncrement, rightIncrement, bottomIncrement);
					pBounds = new Rectangle
						(
						 pBounds.Left - leftIncrement * SubspaceWidth,
						 pBounds.Top - topIncrement * SubspaceHeight,
						 pItems.LengthX * SubspaceWidth,
						 pItems.LengthY * SubspaceHeight
						);
				}
			}
		}

		protected void Cover(Point p)
		{
			Cover(p.X, p.Y);
		}

		protected void Cover(Rectangle r)
		{
			Cover(r.Left, r.Top);
			Cover(r.Right, r.Bottom);
		}

		protected void Insert(Node node)
		{
			Rectangle bounds = node.Bounds;
			Cover(bounds);
			Point lt = SubspaceIndex(bounds.Left, bounds.Top);
			Point rb = SubspaceIndex(bounds.Right, bounds.Bottom);
			for(int x = lt.X; x <= rb.X; x++)
				for(int y = lt.Y; y <= rb.Y; y++)
			{
				Cluster item = (Cluster)pItems[x, y];
				if(item == null)
				{
					item = new Cluster(new Rectangle(Bounds.Left + x * SubspaceWidth, Bounds.Top + y * SubspaceHeight, SubspaceWidth, SubspaceHeight));
					pItems[x, y] = item;
				}
				item.Insert(node);
			}
		}

		protected void Insert(Edge edge)
		{
			Rectangle bounds = edge.Bounds;
			Cover(bounds);
			Point lt = SubspaceIndex(bounds.Left, bounds.Top);
			Point rb = SubspaceIndex(bounds.Right, bounds.Bottom);

			// TODO: use some simple line drawing 
			for(int x = lt.X; x <= rb.X; x++)
				for(int y = lt.Y; y <= rb.Y; y++)
			{
				Cluster item = (Cluster)pItems[x, y];
				if(item == null)
				{
					item = new Cluster(new Rectangle(Bounds.Left + x * SubspaceWidth, Bounds.Top + y * SubspaceHeight, SubspaceWidth, SubspaceHeight));
					pItems[x, y] = item;
				}
				item.Insert(edge);
			}
		}
	}

	public class Space: Quadrant
	{
		protected Graph pGraph;
		public Space(Graph aGraph)
		{
			pGraph = aGraph;
			Cover(pGraph.Bounds);
			foreach(Node node in pGraph.Nodes)
			{
				Insert(node);
			}
			foreach(Edge edge in pGraph.Edges)
			{
				Insert(edge);
			}
		}

		public INodeList SelectNodes(Rectangle clipper, bool orderingMatters)
		{
			if(clipper.Contains(pGraph.Bounds))
			{
				return pGraph.Nodes;
			}
			else if((!IsEmpty) && clipper.IntersectsWith(Bounds))
			{
				Point lt = NormalizedSubspaceIndex(clipper.Left, clipper.Top);
				Point rb = NormalizedSubspaceIndex(clipper.Right, clipper.Bottom);
				NodeCollectionLight result = new NodeCollectionLight();
				try
				{
					for(int x = lt.X; x <= rb.X; x++)
						for(int y = lt.Y; y <= rb.Y; y++)
					{
						Cluster item = (Cluster)pItems[x, y];
						if(item != null)
						{
							foreach(Node node in item.Nodes)
							{
								// to prevent duplicate entries we'll use a marker bit
								if(!node.Mark)
								{
									node.Mark = true;
									result.Add(node);
								}
							}
						}
					}
					if(orderingMatters) result.Sort();
				}
				finally
				{
					foreach(Node node in result)
					{
						node.Mark = false;
					}
				}
				return result;
			}
			else
			{
				return new NodeCollectionLight();
			}
		}

		public EdgeCollectionBase SelectEdges(Rectangle clipper, bool orderingMatters)
		{
			// TODO: it is somehow unreliable
			// return pGraph.Edges;
			if(clipper.Contains(pGraph.Bounds))
			{
				return pGraph.Edges;
			}
			else if((!IsEmpty) && clipper.IntersectsWith(Bounds))
			{
				Point lt = NormalizedSubspaceIndex(clipper.Left, clipper.Top);
				Point rb = NormalizedSubspaceIndex(clipper.Right, clipper.Bottom);
				EdgeCollectionBase result = new EdgeCollectionBase();
				try
				{
					for(int x = lt.X; x <= rb.X; x++)
						for(int y = lt.Y; y <= rb.Y; y++)
					{
						Cluster item = (Cluster)pItems[x, y];
						if(item != null)
						{
							foreach(Edge edge in item.Edges)
							{
								// to prevent duplicate entries we'll use a marker bit
								if(!edge.Mark)
								{
									edge.Mark = true;
									result.Add(edge);
								}
							}
						}
					}
					// TODO: if(orderingMatters) result.Sort();
				}
				finally
				{
					foreach(Edge edge in result)
					{
						edge.Mark = false;
					}
				}
				return result;
			}
			else
			{
				return new EdgeCollectionBase();
			}
		}
	}
}
