using System;

namespace System.IO
{
	/// <summary>
	/// This is simple C# pretty printer capturing TextWriter.WriteLine(string value) calls.
	/// </summary>
	public class CSharpTextWriter: TabbedTextWriter
	{
		private void Setup()
		{
			Lines = new System.Collections.Specialized.StringCollection();
			Tokens = new TokenList(this);
			base.TabSize.OutputMode = Options.Conversion.SpacesToTabs;
		}

		public CSharpTextWriter(string path, bool append, System.Text.Encoding encoding)
			:base(path, append, encoding)
		{
			Setup();
		}

		public CSharpTextWriter(string path): base(path)
		{
			Setup();
		}

		private System.Collections.Specialized.StringCollection Lines;

		#region Tokenizer
		private enum TokenType
		{
			None,
			Other,
			Field,
			Blank,
			Comment,
			BlockBegin,
			BlockEnd
		}

		private class TokenList
		{
			#region Token parser
			private static bool RemoveStartsWith(ref string value, string[] prefix)
			{
				for(int i = 0; i < prefix.Length; i++)
				{
					if (value.StartsWith(prefix[i]))
					{
						value = value.Substring(prefix[i].Length);
						return true;
					}
				}
				return false;
			}

			private static TokenType GetTokenType(string value)
			{
				if (value == null) return TokenType.None;

				string key = value.Trim();
				switch(key)
				{
					case "{":
						return TokenType.BlockBegin;

					case "}":
						return TokenType.BlockEnd;

					case "":
						return TokenType.Blank;

					default:
						if (key.StartsWith("//"))
						{
							return TokenType.Comment;
						}
						else
						{
							if (key.EndsWith(";"))
							{
								int p = key.IndexOf('=');
								if (p >= 0)
								{
									key = key.Substring(0, p);
								}
								if (key.IndexOf('(') < 0)
								{
									if (RemoveStartsWith(ref key, new string[]{"public ", "private ", "protected ", "internal "}))
									{
										if (!key.StartsWith("class "))
										{
											return TokenType.Field;
										}
									}
								}
							}

							return TokenType.Other;
						}
				}
			}
			#endregion

			#region Indentation
			private static int GetCountOfTabs(string line)
			{
				int count = 0;
				for(int i = 0; i < line.Length; i++)
				{
					if (line[i] == '\t')
					{
						count++;
					}
					else
					{
						break;
					}
				}
				return count;
			}
			#endregion

			private System.Collections.Specialized.StringCollection Lines;
			private CSharpTextWriter Owner;

			public TokenList(CSharpTextWriter anOwner)
			{
				Lines = anOwner.Lines;
				Owner = anOwner;
			}

			public int GetLevel()
			{
				return GetLevel(0);
			}

			public int GetLevel(int offset)
			{
				int index = Lines.Count - 1 - offset;
				if ((index >= 0) && (index < Lines.Count))
				{
					return GetCountOfTabs(Lines[index]);
				}
				else
				{
					return 0;
				}
			}

			public TokenType this[int index]
			{
				get
				{
					if ((index >= 0) && (index < Lines.Count))
					{
						return GetTokenType(Lines[index]);
					}
					else
					{
						return TokenType.None;
					}
				}
			}

			public TokenType GetTop()
			{
				return GetTop(0);
			}

			public TokenType GetTop(int offset)
			{
				return this[Lines.Count - 1 - offset];
			}

			public TokenType Push(string value)
			{
				Lines.Add(Owner.Convert(value));

				TokenType result = GetTop(0);
				if (result == TokenType.Blank)
				{
					// normalize
					Lines[Lines.Count - 1] = "";
				}
				return result;
			}

			public void Pop(int offset)
			{
				int index = Lines.Count - 1 - offset;
				if ((index >= 0) && (index < Lines.Count))
				{
					Lines.RemoveAt(index);
				}
			}

			public void Pop()
			{
				Pop(0);
			}
		};

		private TokenList Tokens;
		#endregion

		public override void WriteLine(string value)
		{
			switch(Tokens.Push(value))
			{
				case TokenType.Blank:
					switch(Tokens.GetTop(1))
					{
						case TokenType.None:
						case TokenType.Blank:
						case TokenType.BlockBegin:
							// remove duplicate blank lines or blank lines at
							// the beginning of "{" block or at the beginning
							// of stream
							Tokens.Pop();
							break;
					}
					break;

				case TokenType.BlockEnd:
					switch(Tokens.GetTop(1))
					{
						case TokenType.Blank:
							// remove blank lines before "}"
							Tokens.Pop(1);
							break;
					}
					break;

				case TokenType.Field:
					if (Tokens.GetLevel() > 0)	// ignore base level
					if ((Tokens.GetTop(1) == TokenType.Blank) && (Tokens.GetTop(2) == TokenType.Field))
					{
						// remove blank between fields
						Tokens.Pop(1);
					}
					break;

				case TokenType.Other:
					if (Tokens.GetLevel() > 0)	// ignore base level
					if ((Tokens.GetTop(1) == TokenType.Blank) && (Tokens.GetTop(2) == TokenType.Other))
					{
						// remove blank between other code lines
						Tokens.Pop(1);
					}
					break;
			}
		}

		public override void Close()
		{
			// we have already done the conversion
			base.TabSize.OutputMode = Options.Conversion.None;

			for(int i = 0; i < Lines.Count; i++)
			{
				base.WriteLine(Lines[i]);
			}
			base.Close();
		}
	}
}
