using System;
using System.IO;
using ICSharpCode.SharpZipLib.Zip;
using System.Threading;

namespace System.IO.Compression
{
	/// <summary>
	/// This interface is used by the ZipFile as data source. Names the interface operates on are zip root relative
	/// </summary>
	public interface IFileFeeder
	{
		string[] GetFiles();
		DateTime GetLastWriteTime(string file);
		long GetLength(string file);
		Stream GetInputStream(string file);
		string InternalNameToCompressedName(string file);
		
		long GetChunkSize(string file);
		string GetChunkName(string file, long offset, long length);
	}
	
	public interface IFileWriter
	{
		void CreateDirectory(string folder);
		bool AddFile(string compressedName);
		Stream GetOutputStream(string file, long length);
		string CompressedNameToInternalName(string compressedName);
		void SetLastWriteTime(string file, DateTime stamp);
	}
	
	public class ZipFile: ICompressedFile
	{
		public static bool IsValid(string fileName)
		{
			try
			{
				System.IO.Stream stream = new System.IO.FileStream(fileName, System.IO.FileMode.Open, System.IO.FileAccess.Read);
				try
				{
					try
					{
						int magic1 = stream.ReadByte();
						int magic2 = stream.ReadByte();
						int magic3 = stream.ReadByte();
						int magic4 = stream.ReadByte();
	
						if ((magic1 == 'P') && (magic2 == 'K') && (magic3 == 0x03) && (magic4 == 0x04))
						{
							return true;
						}
					}
					finally
					{
						stream.Close();
					}
				}
				catch(Exception)
				{
				}
			}
			catch(Exception)
			{
			}
			return false;
		}
		
		private IProgress _observer = null;
		public IProgress Observer
		{
			get
			{
				return _observer;
			}
		}
		
		public ZipFile(IProgress observer, string ArchiveFile)
		{
			_observer = observer;
			Open(ArchiveFile, System.IO.FileMode.OpenOrCreate, System.IO.FileAccess.ReadWrite);
		}

		public ZipFile(IProgress observer, string ArchiveFile, System.IO.FileMode mode, System.IO.FileAccess access)
		{
			_observer = observer;
			Open(ArchiveFile, mode, access);
		}
		
		private ZipFile()
		{
		}
		
		public static ZipFile Attach(IProgress observer, string ArchiveFile, string folder)
		{
			ZipFile result = new ZipFile();
			result.pFolder = folder;
			result.pArchiveFile = ArchiveFile;
			result._observer = observer;
			return result;
		}
		
		public ZipFile(string ArchiveFile)
		{
			Open(ArchiveFile, System.IO.FileMode.OpenOrCreate, System.IO.FileAccess.ReadWrite);
		}

		public ZipFile(string ArchiveFile, System.IO.FileMode mode, System.IO.FileAccess access)
		{
			Open(ArchiveFile, mode, access);
		}

		private string pFolder = null;
		public string Folder
		{
			get
			{
				return pFolder;
			}
		}

		private bool pReadOnly = false;
		public bool ReadOnly
		{
			get
			{
				return pReadOnly;
			}
			set
			{
				pReadOnly = value;
			}
		}

		private class FolderWriter: IFileWriter
		{
			private string _folder;
			private string _prefix;
			private ZipFile _handle;
			
			public FolderWriter(ZipFile handle, string folder)
			{
				_handle = handle;
				_folder = folder;
				_prefix = folder;
				if (!_prefix.EndsWith("\\"))
				{
					_prefix = _prefix + "\\";
				}
			}
			
			public void CreateDirectory(string folder)
			{
				System.IO.Directory.CreateDirectory(System.IO.Path.Combine(_folder, folder));
			}
			
			private System.Collections.Generic.List<string> _compressedNames = new System.Collections.Generic.List<string>();
			
			public bool AddFile(string compressedName)
			{
				bool alreadyIn = _compressedNames.Contains(compressedName);
				if (System.IO.File.Exists(System.IO.Path.Combine(_folder, compressedName)))
				{
					if (!alreadyIn)
					{
						// the archive is probably case-sensitive. we need to think out different name for the file
						_handle.AddDuplicateFile(compressedName);
					}
					else
					{
						// this was duplicate entry in the original zip stream
					}
				}
				if (!alreadyIn)
				{
					_compressedNames.Add(compressedName);
				}
				return true;
			}
			
			public Stream GetOutputStream(string file, long length)
			{
				System.IO.Stream result = System.IO.File.Create(System.IO.Path.Combine(_folder, file));
				result.SetLength(length);
				result.Position = 0;
				return result;
			}
			
			public string CompressedNameToInternalName(string compressedName)
			{
				return _handle.CompressedNameToTemporaryName(compressedName).Substring(_prefix.Length);
			}
			
			public void SetLastWriteTime(string file, DateTime stamp)
			{
				System.IO.File.SetLastWriteTime(System.IO.Path.Combine(_folder, file), stamp);
			}
		}
		
		private bool Extract(string ArchiveFile, string Folder, IProgress observer)
		{
			return InternalExtract(ArchiveFile, new FolderWriter(this, Folder), observer);
		}

		private static string CleanZipName(string name)
		{
			if (name.StartsWith("/"))
			{
				return name.Substring(1);
			}
			else
			{
				return name;
			}
		}
		
		private static bool InternalExtract(string ArchiveFile, IFileWriter writer, IProgress observer)
		{
			try
			{
				bool useObserver = observer != null;

				System.IO.Stream input = File.OpenRead(ArchiveFile);
				
				if (useObserver)
				{
					observer.Open(0, (ulong)input.Length);
				}
				
				try
				{
					using (ZipInputStream s = new ZipInputStream(input))
					{
						ZipEntry theEntry;
						while ((theEntry = s.GetNextEntry()) != null)
						{
							string entryName = CleanZipName(theEntry.Name).Replace("/", "\\");
							
							string directoryName;
							string fileName;
							
							if (theEntry.IsDirectory)
							{
								directoryName = entryName;
								fileName = null;
							}
							else
							{
								directoryName = Path.GetDirectoryName(entryName);
								fileName = Path.GetFileName(entryName);
							}
	
							// create directory
							if (!String.IsNullOrEmpty(directoryName))
							{
								// TODO: detect and support case-sensitive directories
								writer.CreateDirectory(directoryName);
							}
	
							if (!String.IsNullOrEmpty(fileName))
							{
								if (!writer.AddFile(entryName))
								{
									continue;
								}
								
								string temporaryFileName = writer.CompressedNameToInternalName(entryName);
								using (Stream streamWriter = writer.GetOutputStream(temporaryFileName, System.Math.Max(theEntry.Size, 0)))
								{
	
									byte[] data = new byte[4096];	// TODO: TUNE:
									while (true)
									{
										int size = s.Read(data, 0, data.Length);
										if (size > 0)
										{
											streamWriter.Write(data, 0, size);
										}
										else
										{
											break;
										}
										
										if (useObserver)
										{
											if (!observer.Step((ulong)input.Position))
											{
												throw new OperationCanceledException();
											}
										}
									}
								}
								
								writer.SetLastWriteTime(temporaryFileName, theEntry.DateTime);
							}
						}
					}
				}
				finally
				{
					if (useObserver)
					{
						observer.Close();
					}
				}
				return true;
			}
			catch (Exception e)
			{
				if (e is OperationCanceledException)
				{
					throw;
				}
				else
				{
					return false;
				}
			}
		}
		
		private string pArchiveFile;
		public void Open(string ArchiveFile, System.IO.FileMode mode, System.IO.FileAccess access)
		{
			pArchiveFile = ArchiveFile;

			ReadOnly = access == System.IO.FileAccess.Read;
			
			pFolder = System.IO.Path.GetTempFileName();
			System.IO.File.Delete(pFolder);
			System.IO.Directory.CreateDirectory(pFolder);

			// TODO: throw exceptions as defined by the FileMode documentation
			switch (mode)
			{
				case FileMode.Create:
				case FileMode.CreateNew:
				case FileMode.Truncate:
					break;

				case FileMode.Append:
				case FileMode.Open:
					if (!Extract(pArchiveFile, pFolder, _observer))
					{
						throw new System.IO.FileNotFoundException();
					}
					break;

				case FileMode.OpenOrCreate:
					Extract(pArchiveFile, pFolder, _observer);
					break;
			}
		}

		#region case-sensitive names
		private struct NameBuffer
		{
			public string Name;
			public string Extension;
			public string Directory;
			
			public NameBuffer(string fileName)
			{
				Name = System.IO.Path.GetFileNameWithoutExtension(fileName);
				Extension = System.IO.Path.GetExtension(fileName);
				if (String.IsNullOrEmpty(Extension))
				{
					Extension = "";
				}
				Directory = System.IO.Path.GetDirectoryName(fileName);
			}
		}
		
		private struct DirectoryName
		{
			public NameBuffer CompressedName;
			public NameBuffer TemporaryName;
		}
		
		private const string TemporaryPrefix = "_$";
		
		private class DirectoryNames: System.Collections.Generic.List<DirectoryName>
		{
			private int pUniqueCounter = 0;
			
			public void AddDuplicate(string compressedName)
			{
				NameBuffer n = new NameBuffer(compressedName);
				
				DirectoryName entry = new DirectoryName();
				entry.CompressedName = n;
				pUniqueCounter++;
				entry.TemporaryName = new NameBuffer(String.Format(TemporaryPrefix + "{0}{1}", pUniqueCounter, n.Extension));
				Add(entry);
			}
			
			public string CompressedNameToTemporaryName(string fileName)
			{
				NameBuffer n = new NameBuffer(fileName);
				foreach(DirectoryName entry in this)
				{
					if ((entry.CompressedName.Name == n.Name) &&
					    true /* TODO: add some support for derived file names??(String.Compare(entry.CompressedName.Extension, n.Extension, true) == 0)*/)
					{
						return System.IO.Path.Combine(entry.CompressedName.Directory, entry.TemporaryName.Name + n.Extension);
					}
				}
				return null;
			}
			
			public string TemporaryNameToCompressedName(string fileName)
			{
				NameBuffer n = new NameBuffer(fileName);
				foreach(DirectoryName entry in this)
				{
					if ((entry.TemporaryName.Name == n.Name) &&
					    true /* TODO: add some support for derived file names??(String.Compare(entry.TemporaryName.Extension, n.Extension, true) == 0)*/)
					{
						return System.IO.Path.Combine(entry.CompressedName.Directory, entry.CompressedName.Name + n.Extension);
					}
				}
				return null;
			}
		}
		
		private class Names: System.Collections.Generic.Dictionary<string, DirectoryNames>
		{
			public DirectoryNames GetDirectory(string fileName, bool canCreate)
			{
				string key = System.IO.Path.GetDirectoryName(fileName).ToUpper();
				DirectoryNames result;
				if (TryGetValue(key, out result))
				{
					return result;
				}
				if (canCreate)
				{
					result = new DirectoryNames();
					Add(key, result);
					return result;
				}
				else
				{
					return null;
				}
			}
		}
		
		private Names pNames = null;
		
		public string CompressedNameToTemporaryName(string fileName)
		{
			if (pNames != null)
			{
				DirectoryNames dir = pNames.GetDirectory(fileName, false);
				if (dir != null)
				{
					string newName = dir.CompressedNameToTemporaryName(fileName);
					if (!String.IsNullOrEmpty(newName))
					{
						return System.IO.Path.Combine(pFolder, newName);
					}
				}
			}
			
			return System.IO.Path.Combine(pFolder, fileName);
		}
		
		public string TemporaryNameToCompressedName(string fileName)
		{
			string result = System.IO.PathBuilder.GetRelativePath(pFolder, fileName);
			
			if (pNames != null)
			{
				DirectoryNames dir = pNames.GetDirectory(result, false);
				if (dir != null)
				{
					string newName = dir.TemporaryNameToCompressedName(result);
					if (!String.IsNullOrEmpty(newName))
					{
						return newName;
					}
				}
			}
			
			return result;
		}
		
		internal void AddDuplicateFile(string compressedName)
		{
			if (pNames == null)
			{
				pNames = new Names();
			}
			DirectoryNames dir = pNames.GetDirectory(compressedName, true);
			dir.AddDuplicate(compressedName);
		}
		
		public string CreateFile(string compressedName)
		{
			string result = CompressedNameToTemporaryName(compressedName);
			if (!System.IO.Path.GetFileName(result).StartsWith(TemporaryPrefix))
			{
				string newKey = System.IO.Path.GetFileNameWithoutExtension(compressedName);
				string folder = System.IO.Path.GetDirectoryName(result);
				if (System.IO.Directory.Exists(folder))
				{
					string[] files = System.IO.Directory.GetFiles(folder, "*.*", System.IO.SearchOption.TopDirectoryOnly);
					foreach(string file in files)
					{
						string key = System.IO.Path.GetFileNameWithoutExtension(TemporaryNameToCompressedName(file));
						if ((String.Compare(newKey, key, true) == 0) && (newKey != key))
						{
							// we have case-sensitivity collision
							AddDuplicateFile(compressedName);
							result = CompressedNameToTemporaryName(compressedName);
							break;
						}
					}
				}
			}
			System.IO.Directory.CreateDirectory(System.IO.Path.GetDirectoryName(result));
			System.IO.File.Create(result).Dispose();
			return result;
		}
		#endregion

		#region Static helpers
		public static void Extract(IProgress observer, string archiveFile, string targetFolder)
		{
			targetFolder = System.IO.Path.GetFullPath(targetFolder);
			if (!System.IO.Directory.Exists(targetFolder))
			{
				throw new System.IO.DirectoryNotFoundException();
			}
			ZipFile zip = new ZipFile();
			try
			{
				zip.pFolder = targetFolder;
				if (!zip.Extract(archiveFile, targetFolder, observer))
				{
					throw new System.IO.FileNotFoundException();
				}
			}
			finally
			{
				zip.Detach();
				zip.Dispose();
			}
		}
		
		public static void Extract(IProgress observer, string archiveFile, IFileWriter writer)
		{
			if (!InternalExtract(archiveFile, writer, observer))
			{
				throw new System.IO.FileNotFoundException();
			}
		}
		
		public static void Extract(string archiveFile, IFileWriter writer)
		{
			Extract(null, archiveFile, writer);
		}
		
		public static void Extract(string archiveFile, string targetFolder)
		{
			Extract(null, archiveFile, targetFolder);
		}
		
		public static void Create(string archiveFile, string sourceFolder, string[] sourceFiles)
		{
			InternalCreate(archiveFile, new FolderFeeder(sourceFolder, sourceFiles), null);
		}
		
		public static void Create(IProgress observer, string archiveFile, string sourceFolder, string[] sourceFiles)
		{
			InternalCreate(archiveFile, new FolderFeeder(sourceFolder, sourceFiles), observer);
		}

		public static void Create(IProgress observer, string archiveFile, string sourceFolder)
		{
			InternalCreate(archiveFile, new FolderFeeder(sourceFolder), observer);
		}
		
		public static void Create(string archiveFile, string sourceFolder)
		{
			InternalCreate(archiveFile, new FolderFeeder(sourceFolder), null);
		}
		
		public static void Create(IProgress observer, string archiveFile, IFileFeeder feeder)
		{
			InternalCreate(archiveFile, feeder, observer);
		}
			
		public static void Create(string archiveFile, IFileFeeder feeder)
		{
			InternalCreate(archiveFile, feeder, null);
		}
		#endregion
		
		private class FolderFeeder: IFileFeeder
		{
			private string _folder;
			private string _prefix;
			private string[] _files;
			
			public FolderFeeder(string folder)
			{
				Initialize(folder, null);
			}
			
			public FolderFeeder(string folder, string[] files)
			{
				Initialize(folder, files);
			}
			
			private void Initialize(string folder, string[] files)
			{
				_folder = System.IO.Path.GetFullPath(folder);
				_prefix = _folder;
				if (!_prefix.EndsWith("\\"))
				{
					_prefix = _prefix + "\\";
				}
				_files = files;
			}
			
			protected string InnerToOuter(string fileName)
			{
				return fileName.Substring(_prefix.Length);
			}
			
			protected string OuterToInner(string fileName)
			{
				return _prefix + fileName;
			}
			
			public string[] GetFiles()
			{
				if (_files == null)
				{
					_files = Directory.GetFiles(_folder, "*.*", SearchOption.AllDirectories);
					for(int i = 0; i < _files.Length; i++)
					{
						_files[i] = InnerToOuter(_files[i]);
					}
				}
				return _files;
			}
			
			public DateTime GetLastWriteTime(string file)
			{
				return File.GetLastAccessTime(OuterToInner(file));
			}
			
			public long GetLength(string file)
			{
				return (new FileInfo(OuterToInner(file))).Length;
			}
			
			public Stream GetInputStream(string file)
			{
				return File.OpenRead(OuterToInner(file));
			}
			
			public virtual string InternalNameToCompressedName(string file)
			{
				return file;
			}
			
			public long GetChunkSize(string file)
			{
				return 0;	
			}
			
			public string GetChunkName(string file, long offset, long length)
			{
				throw new NotSupportedException();
			}
		}
		
		private class TemporaryFolderFeeder: FolderFeeder
		{
			private ZipFile _zip;
			public TemporaryFolderFeeder(ZipFile zip): base(zip.Folder)
			{
				_zip = zip;
			}
			
			public override string InternalNameToCompressedName(string file)
			{
				return _zip.TemporaryNameToCompressedName(OuterToInner(file));
			}
		}
		
		private static void InternalCreate(string archiveFile, IFileFeeder feeder, IProgress observer)
		{
			string pArchiveFile = System.IO.Path.GetFullPath(archiveFile);
			
			// Depending on the directory this could be very large and would require more attention
			// in a commercial package.
			string[] filenames = feeder.GetFiles();

			bool useObserver = observer != null;
			long step = 0;
			if (useObserver)
			{
				// calculate the progress's max
				long sum = 0;
				foreach(string file in filenames)
				{
					sum += feeder.GetLength(file);
				}
				observer.Open(0, (ulong)sum);
			}
			
			try
			{
				// 'using' statements gaurantee the stream is closed properly which is a big source
				// of problems otherwise.  Its exception safe as well which is great.
				using (ZipOutputStream s = new ZipOutputStream(File.Create(pArchiveFile)))
				{
					s.SetLevel(9); // 0 - store only to 9 - means best compression
					s.UseZip64 = ICSharpCode.SharpZipLib.Zip.UseZip64.Off;
	
					// Using a fixed size buffer here makes no noticeable difference for output
					// but keeps a lid on memory usage.
					byte[] buffer = new byte[4096];
	
					foreach (string file in filenames)
					{
						string fileName = feeder.InternalNameToCompressedName(file);
						
						using (Stream fs = feeder.GetInputStream(file))
						{
							long fileSize = fs.Length;
							long chunkSize = feeder.GetChunkSize(file);
							bool useChunks = (chunkSize > 0) && (chunkSize < fileSize);
							
							if (useChunks)
							{
								long offset = 0;
								long size;
								long toGo = fileSize;
								
								while (toGo > 0)
								{
									size = chunkSize;
									if (size > toGo)
									{
										size = toGo;
									}
									ZipEntry entry = new ZipEntry(feeder.GetChunkName(fileName, offset, size).Replace("\\", "/"));
									
									entry.DateTime = feeder.GetLastWriteTime(file);
									entry.Size = size;
									s.PutNextEntry(entry);
									
									offset += size;
									toGo -= size;
									
									try
									{
										int sourceBytes;
										do
										{
											int bufferLength = buffer.Length;
											if (bufferLength > size)
											{
												bufferLength = (int)size;
											}
											if (bufferLength == 0)
											{
												break;
											}
											sourceBytes = fs.Read(buffer, 0, bufferLength);
											s.Write(buffer, 0, sourceBytes);
											
											if (useObserver)
											{
												step += sourceBytes;
												if (!observer.Step((ulong)step))
												{
													throw new OperationCanceledException();
												}
											}
											
											size -= sourceBytes;
										}
										while (sourceBytes > 0);
									}
									catch(Exception)
									{
										// sort of "rollback" the entry
										entry.Size = -1;
										throw;
									}
								}
							}
							else
							{
								ZipEntry entry = new ZipEntry(fileName.Replace("\\", "/"));
								
								entry.DateTime = feeder.GetLastWriteTime(file);
								entry.Size = fileSize;
								s.PutNextEntry(entry);
		
								try
								{
									int sourceBytes;
									do
									{
										sourceBytes = fs.Read(buffer, 0, buffer.Length);
										s.Write(buffer, 0, sourceBytes);
										
										if (useObserver)
										{
											step += sourceBytes;
											if (!observer.Step((ulong)step))
											{
												throw new OperationCanceledException();
											}
										}
									}
									while (sourceBytes > 0);
								}
								catch(Exception)
								{
									// sort of "rollback" the entry
									entry.Size = -1;
									throw;
								}
							}
						}
					}
	
					// Finish/Close arent needed strictly as the using statement does this automatically
	
					// Finish is important to ensure trailing information for a Zip file is appended.  Without this
					// the created file would be invalid.
					s.Finish();
	
					// Close is important to wrap things up and unlock the file.
					s.Close();
				}
			}
			finally
			{
				if (useObserver)
				{
					observer.Close();
				}
			}
		}
		
		public void Detach()
		{
			ReadOnly = true;
			pFolder = null;
		}
		
		private void Close()
		{
			if (!ReadOnly)
			{
				InternalCreate(pArchiveFile, new TemporaryFolderFeeder(this), _observer);
			}
			if (pFolder != null)
			{
				System.IO.Directory.Delete(pFolder, true);
			}

			pFolder = null;
		}

		#region Implementation of IDisposable
		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		private void Dispose(bool disposing)
		{
			if (!disposing)
			{
				this.ReadOnly = true;
			}
				
			if (pFolder != null)
			{
				Close();
			}
		}

		~ZipFile()
		{
			Dispose(false);
		}
		#endregion
	}
}
