using System;
using System.Text.RegularExpressions;

namespace System.Text
{
	public class StringFilter
	{
		public static string WildcardToRegex(string pattern)
		{
			return "^" + Regex.Escape(pattern).
			Replace("\\*", ".*").
			Replace("\\?", ".") + "$";
		}
		
		private class MaskList: System.Collections.Generic.List<Regex>
		{
			private System.Collections.Generic.List<string> _masks = new System.Collections.Generic.List<string>();
			
			private bool _caseSensitive;
			public MaskList(bool caseSensitive)
			{
				_caseSensitive = caseSensitive;
			}
			
			public void Add(string mask)
			{
				_masks.Add(mask);
				RegexOptions options = RegexOptions.Singleline | RegexOptions.Compiled;
				if (!_caseSensitive)
				{
					options = options | RegexOptions.IgnoreCase;
				}
				Regex item = new Regex(WildcardToRegex(mask), options);
				base.Add(item);
			}
			
			public string[] Masks
			{
				get
				{
					return _masks.ToArray();
				}
			}
		}
		
		private void Initialize(bool caseSensitive)
		{
			_includeMasks = new MaskList(caseSensitive);
			_excludeMasks = new MaskList(caseSensitive);
		}
		
		private MaskList _includeMasks;
		private MaskList _excludeMasks;
		
		public StringFilter(bool caseSensitive)
		{
			Initialize(caseSensitive);
		}
		
		public StringFilter(string masks, bool caseSensitive)
		{
			Initialize(caseSensitive);
			Add(masks);
		}
		
		protected string[] IncludeMasks
		{
			get
			{
				return _includeMasks.Masks;
			}
		}
		
		protected string[] ExcludeMasks
		{
			get
			{
				return _excludeMasks.Masks;
			}
		}
		
		/// <summary>
		/// Adds wildcarded include or exclude masks. Accepted wildcard characters:
		/// * - 0 or more characters of any value
		/// ? - 1 character of any value
		/// </summary>
		/// <param name="masks">semicolon (or comma) separated list of masks. Each mask can start with + or - sign indicating "include" or "exclude" category</param>
		public void Add(string masks)
		{
			foreach(string item in masks.Replace(";", ",").Split(','))
			{
				string mask = item.Trim();
				if (mask.StartsWith("-"))
				{
					Exclude(mask.Substring(1));
				}
				else if (mask.StartsWith("+"))
				{
					Include(mask.Substring(1));
				}
				else
				{
					Include(mask);
				}
			}
		}
		
		protected virtual string NormalizedMask(string mask)
		{
			return mask;
		}
		
		protected virtual string NormalizedText(string text)
		{
			return text;
		}
		
		protected virtual void AddIncludeMask(string mask)
		{
			_includeMasks.Add(mask);
		}
		
		protected virtual void AddExcludeMask(string mask)
		{
			_excludeMasks.Add(mask);
		}
		
		public void Include(string mask)
		{
			mask = NormalizedMask(mask);
			if (!String.IsNullOrEmpty(mask))
			{
				AddIncludeMask(mask);
			}
		}
		
		public void Exclude(string mask)
		{
			mask = NormalizedMask(mask);
			if (!String.IsNullOrEmpty(mask))
			{
				AddExcludeMask(mask);
			}
		}
		
		private static bool Match(string text, MaskList masks)
		{
			foreach(Regex mask in masks)
			{
				if (mask.IsMatch(text))
				{
					return true;
				}
			}
			return false;
		}
		
		public virtual bool Accepts(string text)
		{
			text = NormalizedText(text);
			if (Match(text, _excludeMasks))
			{
				return false;
			}
			if ((_includeMasks.Count == 0) || Match(text, _includeMasks))
			{
				return true;
			}
			return false;
		}
		
		public virtual bool AcceptsAll()
		{
			return (_excludeMasks.Count == 0) && (_includeMasks.Count == 0);
		}
	}
}
