using System;

namespace System.Windows.Forms
{
	public class MessageBoardClient: IDisposable
	{
		public delegate void MessageProcessor(System.Threading.MessageBoard.Entry message);
		
		private System.Windows.Forms.Timer _ticker = null;
		
		public MessageBoardClient(Guid id)
		{
			Initialize(id, null);
		}
		
		public MessageBoardClient(Guid id, MessageProcessor processor)
		{
			Initialize(id, processor);
		}
		
		private MessageProcessor _processMessage = null;
		public MessageProcessor ProcessMessage
		{
			get
			{
				return _processMessage;
			}
			set
			{
				if (_processMessage != null)
				{
					_ticker.Enabled = false;
				}
				_processMessage = value;
				if ((_processMessage != null) && Enabled)
				{
					_ticker.Enabled = true;
				}
			}
		}
		
		private bool _enabled = true;
		public bool Enabled
		{
			get
			{
				return _enabled;
			}
			set
			{
				if (value)
				{
					_enabled = true;
					if (_processMessage != null)
					{
						_ticker.Enabled = true;
					}
				}
				else
				{
					_enabled = false;
					_ticker.Enabled = false;
				}
			}
		}
		
		private void Initialize(Guid id, MessageProcessor processor)
		{
			_board = new System.Threading.MessageBoard(id);
			_ticker = new System.Windows.Forms.Timer();
			_ticker.Interval = 1000; 	// TODO: TUNE: polling frequency
			_ticker.Enabled = false;
			_ticker.Tick += this.TryReceiveMessage;
			ProcessMessage = processor;
		}
		
		private bool InternalProcessMessage()
		{
			System.Threading.MessageBoard.Entry message = _board.Receive();
			if (message == null)
			{
				return false;
			}
			else
			{
				try
				{
					if (_processMessage != null)
					{
						_processMessage(message);
					}
				}
				finally
				{
					message.Dispose();
				}
				return true;
			}
		}
		
		public void ReceiveAll()
		{
			while (InternalProcessMessage())
			{
			}
		}
		
		private void TryReceiveMessage(object sender, EventArgs e)
		{
			_ticker.Enabled = false;
			try
			{
				ReceiveAll();
			}
			finally
			{
				_ticker.Enabled = true;
			}
		}
		
		public void Send(string to, string subject, string body)
		{
			_board.Send(to, subject, body);
		}

		public void Broadcast(string subject, string body)
		{
			_board.Send(null, subject, body);
		}
		
		private System.Threading.MessageBoard _board;
		protected System.Threading.MessageBoard Board
		{
			get
			{
				return _board;				
			}
		}
		
		#region IDisposable interface implementation
		~MessageBoardClient()
		{
			Dispose(false);
		}

		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		private bool _disposed = false;
		private void Dispose(bool commit)
		{
			if (!_disposed)
			{
				_disposed = true;
				if (_ticker != null)
				{
					_ticker.Dispose();
					_ticker = null;
				}
				if (_board != null)
				{
					_board.Dispose();
					_board = null;
				}
			}
		}
		#endregion
	}
}
